"""
This script shows an example of how to use Ned's vision to
make a conditioning with any of the objects supplied with the Vision Kit
The script works in 2 ways:
- One where all the vision process is made on the robot
- One where the vision process is made on the computer

The first one shows how easy it is to use Ned's vision kit with PyNiryo2
The second demonstrates a way to do image processing from user's computer. It highlights
the fact that the user can imagine every type of process on his computer.

The objects will be conditioned in a grid of dimension grid_dimension. If the grid is completed,
objects will be pack over the lower level
"""
from pyniryo2 import *
import pyniryo.vision as vision
import cv2

# -- MUST Change these variables
simulation_mode = True
if simulation_mode:
    robot_ip_address, workspace_name = "127.0.0.1", "gazebo_1"
else:
    robot_ip_address, workspace_name = "10.10.10.10", "workspace_1"

# -- Can Change these variables
grid_dimension = (3, 3)  # conditioning grid dimension
vision_process_on_robot = False  # boolean to indicate if the image processing append on the Robot
display_stream = True  # Only used if vision on computer

# -- Should Change these variables
# The pose from where the image processing happens
observation_pose = PoseObject(
    x=0.17, y=0., z=0.35,
    roll=0.0, pitch=1.57, yaw=0.0,
)

# Center of the conditioning area
center_conditioning_pose = PoseObject(
    x=0.0, y=-0.25, z=0.12,
    roll=-0., pitch=1.57, yaw=-1.57
)


# -- MAIN PROGRAM

def process(niryo_robot):
    """

    :type niryo_robot: NiryoRobot
    :rtype: None
    """
    # Initializing variables
    obj_pose = None
    try_without_success = 0
    count = 0
    if not vision_process_on_robot:
        camera_info = niryo_robot.vision.get_camera_intrinsics()
        mtx, dist = camera_info.intrinsics, camera_info.distortion
    else:
        mtx = dist = None
    # Loop
    while try_without_success < 5:
        # Moving to observation pose
        niryo_robot.arm.move_pose(observation_pose)

        if vision_process_on_robot:
            ret = niryo_robot.vision.get_target_pose_from_cam(workspace_name,
                                                              height_offset=0.0,
                                                              shape=ObjectShape.ANY,
                                                              color=ObjectColor.ANY)
            obj_found, obj_pose, shape, color = ret

        else:  # Home made image processing
            img_compressed = niryo_robot.vision.get_img_compressed()
            img = vision.uncompress_image(img_compressed)
            img = vision.undistort_image(img, mtx, dist)
            # extracting working area
            im_work = vision.extract_img_workspace(img, workspace_ratio=1.0)
            if im_work is None:
                print("Unable to find markers")
                try_without_success += 1
                if display_stream:
                    cv2.imshow("Last image saw", img)
                    cv2.waitKey(25)
                continue

            # Applying Threshold on ObjectColor
            color_hsv_setting = vision.ColorHSV.ANY.value
            img_thresh = vision.threshold_hsv(im_work, *color_hsv_setting)

            if display_stream:
                vision.show_img("Last image saw", img, wait_ms=100)
                vision.show_img("Image thresh", img_thresh, wait_ms=100)
            # Getting biggest contour/blob from threshold image
            contour = vision.biggest_contour_finder(img_thresh)
            if contour is None or len(contour) == 0:
                print("No blob found")
                obj_found = False
            else:
                img_thresh_rgb_w_contour = vision.draw_contours(img_thresh, [contour])
                vision.show_img("Image thresh", img_thresh_rgb_w_contour, wait_ms=30)

                # Getting contour/blob center and angle
                cx, cy = vision.get_contour_barycenter(contour)
                cx_rel, cy_rel = vision.relative_pos_from_pixels(im_work, cx, cy)
                angle = vision.get_contour_angle(contour)

                # Getting object world pose from relative pose
                obj_pose = niryo_robot.vision.get_target_pose_from_rel(workspace_name,
                                                                       height_offset=0.0,
                                                                       x_rel=cx_rel, y_rel=cy_rel,
                                                                       yaw_rel=angle)
                obj_found = True
        if not obj_found:
            try_without_success += 1
            continue
        # Everything is good, so we going to object
        niryo_robot.pick_place.pick_from_pose(obj_pose)

        # Computing new place pose
        offset_x = count % grid_dimension[0] - grid_dimension[0] // 2
        offset_y = (count // grid_dimension[1]) % 3 - grid_dimension[1] // 2
        offset_z = count // (grid_dimension[0] * grid_dimension[1])
        place_pose = center_conditioning_pose.copy_with_offsets(0.05 * offset_x, 0.05 * offset_y, 0.025 * offset_z)

        # Placing
        niryo_robot.pick_place.place_from_pose(place_pose)

        try_without_success = 0
        count += 1


if __name__ == '__main__':
    # Connect to robot
    robot = NiryoRobot(robot_ip_address)
    # Changing tool
    robot.tool.update_tool()
    # Calibrate robot if robot needs calibration
    robot.arm.calibrate_auto()
    # Launching main process
    process(robot)
    # Ending
    robot.arm.go_to_sleep()
    # Releasing connection
    robot.end()



#!/usr/bin/env python
# coding=utf-8


class HardwareStatusObject:
    """
    Object used to store every hardware information

    :ivar rpi_temperature: Number representing the rpi temperature
    :vartype rpi_temperature: float
    :ivar hardware_version: Number representing the hardware version
    :vartype hardware_version: str
    :ivar connection_up: Boolean indicating if the connection with the robot is up
    :vartype connection_up: bool
    :ivar error_message: Error message status on error
    :vartype error_message: str
    :ivar calibration_needed: Boolean indicating if a calibration is needed
    :vartype calibration_needed: bool
    :ivar calibration_in_progress: Boolean indicating if calibration is in progress
    :vartype calibration_in_progress: bool
    :ivar motor_names: List of motor names
    :vartype motor_names: list[str]
    :ivar motor_types: List of motor types
    :vartype motor_types: list[str]
    :ivar motors_temperature: List of motors_temperature
    :vartype motors_temperature: list[float]
    :ivar motors_voltage: List of motors_voltage
    :vartype motors_voltage: list[float]
    :ivar hardware_errors: List of hardware errors
    :vartype hardware_errors: list[int]
    :ivar hardware_error_messages: List of hardware error messages
    :vartype hardware_error_messages: list[str]

    """

    def __init__(self):
        self.rpi_temperature = None
        self.hardware_version = None
        self.connection_up = None
        self.error_message = None
        self.calibration_needed = None
        self.calibration_in_progress = None

        # Following list describe each motor
        # Row 0 for first motor, row 1 for second motor, row 2 for third motor, row 3 for fourth motor
        self.motor_names = None
        self.motor_types = None
        self.motors_temperature = None
        self.motors_voltage = None
        self.hardware_errors = None
        self.hardware_error_messages = None

    def init_from_message(self, msg):
        # Number representing the rpi temperature
        self.rpi_temperature = msg["rpi_temperature"]
        # Number representing the hardware version
        self.hardware_version = msg["hardware_version"]
        # Boolean indicating if the connection with the robot is up
        self.connection_up = msg["connection_up"]
        # Error message status on error
        self.error_message = msg["error_message"]
        # Boolean indicating if a calibration is needed
        self.calibration_needed = msg["calibration_needed"]
        # Boolean indicating if calibration is in progress
        self.calibration_in_progress = msg["calibration_in_progress"]

        # Following list describe each motor
        # Row 0 for first motor, row 1 for second motor, row 2 for third motor, row 3 for fourth motor
        # List of motor names
        self.motor_names = msg["motor_names"]
        # List of motor types
        self.motor_types = msg["motor_types"]
        # List of motors_temperature
        self.motors_temperature = msg["temperatures"]
        # List of motors_voltage
        self.motors_voltage = msg["voltages"]
        # List of hardware errors
        self.hardware_errors = msg["hardware_errors"]
        # List of hardware error messages
        self.hardware_error_messages = msg["hardware_errors_message"]

    def init_from_values(self, rpi_temperature, hardware_version, connection_up,
                         error_message, calibration_needed, calibration_in_progress,
                         motor_names, motor_types,
                         motors_temperature, motors_voltage, hardware_errors, hardware_error_messages):
        # Number representing the rpi temperature
        self.rpi_temperature = rpi_temperature
        # Number representing the hardware version
        self.hardware_version = hardware_version
        # Boolean indicating if the connection with the robot is up
        self.connection_up = connection_up
        # Error message status on error
        self.error_message = error_message
        # Boolean indicating if a calibration is needed
        self.calibration_needed = calibration_needed
        # Boolean indicating if calibration is in progress
        self.calibration_in_progress = calibration_in_progress

        # Following list describe each motor
        # Row 0 for first motor, row 1 for second motor, row 2 for third motor, row 3 for fourth motor
        # List of motor names
        self.motor_names = motor_names
        # List of motor types
        self.motor_types = motor_types
        # List of motors_temperature
        self.motors_temperature = motors_temperature
        # List of motors_voltage
        self.motors_voltage = motors_voltage
        # List of hardware errors
        self.hardware_errors = hardware_errors
        # List of hardware error messages
        self.hardware_error_messages = hardware_error_messages

    def __str__(self):
        list_string_ret = list()
        list_string_ret.append("Temp (°C) : {}".format(self.rpi_temperature))
        list_string_ret.append("Hardware version : {}".format(self.hardware_version))
        list_string_ret.append("Connection Up : {}".format(self.connection_up))
        list_string_ret.append("Error Message : {}".format(self.error_message))
        list_string_ret.append("Calibration Needed : {}".format(self.calibration_needed))
        list_string_ret.append("Calibration in progress : {}".format(self.calibration_in_progress))
        list_string_ret.append("MOTORS INFOS : Motor1, Motor2, Motor3, Motor4, Motor5, Motor6,")
        list_string_ret.append("Names : {}".format(self.motor_names))
        list_string_ret.append("Types : {}".format(self.motor_types))
        list_string_ret.append("Temperatures : {}".format(self.motors_temperature))
        list_string_ret.append("Voltages : {}".format(self.motors_voltage))
        list_string_ret.append("Hardware errors : {}".format(self.hardware_errors))
        list_string_ret.append("Hardware error messages : {}".format(self.hardware_error_messages))
        return "\n".join(list_string_ret)

    def __repr__(self):
        return self.__str__()


class JointStateObject:
    """
    Object used to store every joint state information

    :ivar name: List of joint names
    :vartype name: list[str]
    :ivar position: List of joint positions
    :vartype position: list[float]
    :ivar velocity: List of joint velocities
    :vartype velocity: list[float]
    :ivar effort: List of joint efforts
    :vartype effort: list[float]
    """

    def __init__(self):
        # Following list describe each joint
        # Row 0 for first joint, row 1 for second joint, row 2 for third joint, row 3 for fourth joint, ...

        # List of joint names
        self.name = None
        # List of joint positions
        self.position = None
        # List of joint velocities
        self.velocity = None
        # List of joint efforts
        self.effort = None

    def init_from_message(self, msg):
        # Following list describe each joint
        # Row 0 for first joint, row 1 for second joint, row 2 for third joint, row 3 for fourth joint, ...

        # List of joint names
        self.name = msg["name"]
        # List of joint positions
        self.position = msg["position"]
        # List of joint velocities
        self.velocity = msg["velocity"]
        # List of joint efforts
        self.effort = msg["effort"]

    def init_from_values(self, names, positions, velocities, efforts):
        # Following list describe each joint
        # Row 0 for first joint, row 1 for second joint, row 2 for third joint, row 3 for fourth joint, ...

        # List of joint names
        self.name = names
        # List of joint positions
        self.position = positions
        # List of joint velocities
        self.velocity = velocities
        # List of joint efforts
        self.effort = efforts

    def __str__(self):
        list_string_ret = list()
        list_string_ret.append("MOTORS INFOS : ")
        list_string_ret.append("Names : {}".format(self.name))
        list_string_ret.append("Positions : {}".format(self.position))
        list_string_ret.append("Velocities : {}".format(self.velocity))
        list_string_ret.append("Efforts : {}".format(self.effort))
        return "\n".join(list_string_ret)

    def __repr__(self):
        return self.__str__()



import roslibpy.actionlib

from pyniryo2.enums import ArmMoveCommandType


class TrajectoriesActions(object):

    def __init__(self, client):
        self.__client = client

        self.trajectory_action = None
        self.trajectory_action = roslibpy.actionlib.ActionClient(self.__client,
                                                                 '/niryo_robot_arm_commander/robot_action/',
                                                                 'niryo_robot_arm_commander/RobotMoveAction')

    def __del__(self):
        if self.trajectory_action is not None:
            self.trajectory_action.cancel()
            self.trajectory_action.dispose()

    def get_execute_trajectories_goal(self, pose_list, dist_smoothing=0):
        cmd = {'cmd_type': ArmMoveCommandType.EXECUTE_TRAJ.value,
               'list_poses': [self.pose_quat_list_to_dict(pose) for pose in pose_list],
               'dist_smoothing': dist_smoothing}
        return roslibpy.actionlib.Goal(self.trajectory_action, roslibpy.Message({'cmd': cmd}))

    @staticmethod
    def pose_quat_list_to_dict(pose_list):
        return {"position": dict(zip(["x", "y", "z"], pose_list[:3])),
                "orientation": dict(zip(["x", "y", "z", "w"], pose_list[3:]))}



#!/usr/bin/env python
# coding=utf-8

from time import time
import numpy as np


class PoseObject:
    """
    Pose object which stores x, y, z, roll, pitch & yaw parameters

    :ivar x: X (meter)
    :vartype x: float
    :ivar y: Y (meter)
    :vartype y: float
    :ivar z: Z (meter)
    :vartype z: float
    :ivar roll: Roll (radian)
    :vartype roll: float
    :ivar pitch: Pitch (radian)
    :vartype pitch: float
    :ivar yaw: Yaw (radian)
    :vartype yaw: float

    """

    def __init__(self, x, y, z, roll, pitch, yaw):
        # X (meter)
        self.x = float(x)
        # Y (meter)
        self.y = float(y)
        # Z (meter)
        self.z = float(z)
        # Roll (radian)
        self.roll = float(roll)
        # Pitch (radian)
        self.pitch = float(pitch)
        # Yaw (radian)
        self.yaw = float(yaw)

    def __str__(self):
        position = "x = {:.4f}, y = {:.4f}, z = {:.4f}".format(self.x, self.y, self.z)
        orientation = "roll = {:.3f}, pitch = {:.3f}, yaw = {:.3f}".format(self.roll, self.pitch, self.yaw)
        return position + "\n" + orientation

    def __repr__(self):
        return self.__str__()

    def __add__(self, other):
        x = self.x + other.x
        y = self.y + other.y
        z = self.z + other.z
        roll = self.roll + other.roll
        pitch = self.pitch + other.pitch
        yaw = self.yaw + other.yaw
        return PoseObject(x, y, z, roll, pitch, yaw)

    def __sub__(self, other):
        x = self.x - other.x
        y = self.y - other.y
        z = self.z - other.z
        roll = self.roll - other.roll
        pitch = self.pitch - other.pitch
        yaw = self.yaw - other.yaw
        return PoseObject(x, y, z, roll, pitch, yaw)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y and self.z == other.z \
               and self.roll == other.roll and self.pitch == other.pitch and self.yaw == other.yaw

    def copy_with_offsets(self, x_offset=0., y_offset=0., z_offset=0., roll_offset=0., pitch_offset=0., yaw_offset=0.):
        """
        Create a new pose from copying from copying actual pose with offsets

        :rtype: PoseObject
        """
        return PoseObject(self.x + x_offset,
                          self.y + y_offset,
                          self.z + z_offset,
                          self.roll + roll_offset,
                          self.pitch + pitch_offset,
                          self.yaw + yaw_offset)

    def to_list(self):
        """
        Return a list [x, y, z, roll, pitch, yaw] corresponding to the pose's parameters

        :rtype: list[float]
        """
        list_pos = [self.x, self.y, self.z, self.roll, self.pitch, self.yaw]
        return list(map(float, list_pos))

    @property
    def quaternion(self):
        """
        Return the quaternion in a list [qx, qy, qz, qw]

        :return: quaternion [qx, qy, qz, qw]
        :rtype: list[float, float, float, float]
        """
        return self.euler_to_quaternion(self.roll, self.pitch, self.yaw)

    @property
    def quaternion_pose(self):
        """
        Return the position and the quaternion in a list [x, y, z, qx, qy, qz, qw]

        :return: position [x, y, z] + quaternion [qx, qy, qz, qw]
        :rtype: list[float, float, float, float, float, float, float]

        """
        return [self.x, self.y, self.z] + list(self.euler_to_quaternion(self.roll, self.pitch, self.yaw))

    @staticmethod
    def euler_to_quaternion(roll, pitch, yaw):
        """
        Convert euler angles to quaternion

        :param roll: roll in radians
        :type roll: float
        :param pitch: pitch in radians
        :type pitch: float
        :param yaw: yaw in radians
        :type yaw: float
        :return: quaternion in a list [qx, qy, qz, qw]
        :rtype: list[float, float, float, float]
        """
        qx = np.sin(roll / 2) * np.cos(pitch / 2) * np.cos(yaw / 2) - np.cos(roll / 2) * np.sin(pitch / 2) * np.sin(
            yaw / 2)
        qy = np.cos(roll / 2) * np.sin(pitch / 2) * np.cos(yaw / 2) + np.sin(roll / 2) * np.cos(pitch / 2) * np.sin(
            yaw / 2)
        qz = np.cos(roll / 2) * np.cos(pitch / 2) * np.sin(yaw / 2) - np.sin(roll / 2) * np.sin(pitch / 2) * np.cos(
            yaw / 2)
        qw = np.cos(roll / 2) * np.cos(pitch / 2) * np.cos(yaw / 2) + np.sin(roll / 2) * np.sin(pitch / 2) * np.sin(
            yaw / 2)

        return [qx, qy, qz, qw]

    @staticmethod
    def quaternion_to_euler_angle(qx, qy, qz, qw):
        """
        Convert euler angles to quaternion

        :param qx:
        :type qx: float
        :param qy:
        :type qy: float
        :param qz:
        :type qz: float
        :param qw:
        :type qw: float
        :return: euler angles in a list [roll, pitch, yaw]
        :rtype: list[float, float, float]
        """
        ysqr = qy * qy

        t0 = +2.0 * (qw * qx + qy * qz)
        t1 = +1.0 - 2.0 * (qx * qx + ysqr)
        roll = np.arctan2(t0, t1)

        t2 = +2.0 * (qw * qy - qz * qx)

        t2 = np.clip(t2, a_min=-1.0, a_max=1.0)
        pitch = np.arcsin(t2)

        t3 = +2.0 * (qw * qz + qx * qy)
        t4 = +1.0 - 2.0 * (ysqr + qz * qz)
        yaw = np.arctan2(t3, t4)

        return roll, pitch, yaw


class TrajectoryObject():

    def __init__(self, name, description, joint_names, points, accelerations, velocities, time_from_start):
        self.name = name
        self.description = description
        self.joint_names = joint_names
        self.points = points
        self.accelerations = accelerations
        self.velocities = velocities
        self.time_from_start = time_from_start



import numpy as np
import base64

from pyniryo2.niryo_topic import NiryoTopic
from .objects import CameraInfo, ImageParameters


class VisionTopics(object):

    def __init__(self, client):
        self.__client = client

        self.compressed_video_stream_topic = NiryoTopic(self.__client,
                                                        '/niryo_robot_vision/compressed_video_stream',
                                                        'sensor_msgs/CompressedImage',
                                                        compressed_video_stream_topic_conversion)

        self.camera_info_topic = NiryoTopic(self.__client,
                                            '/niryo_robot_vision/camera_intrinsics',
                                            'sensor_msgs/CameraInfo',
                                            camera_info_topic_conversion)

        self.video_stream_parameters_topic = NiryoTopic(self.__client,
                                                        '/niryo_robot_vision/video_stream_parameters',
                                                        'niryo_robot_vision/ImageParameters',
                                                        video_stream_parameters_topic_conversion)


def compressed_video_stream_topic_conversion(msg):
    # Convert base64 into uint8 array
    return base64.b64decode(msg['data'])


def camera_info_topic_conversion(msg):
    mtx = np.reshape(msg['K'], (3, 3))
    dist = np.expand_dims(msg['D'], axis=0)

    return CameraInfo(intrinsics=mtx, distortion=dist)


def video_stream_parameters_topic_conversion(msg):
    return ImageParameters(brightness_factor=msg["brightness_factor"], contrast_factor=msg["contrast_factor"],
                           saturation_factor=msg["saturation_factor"])



# - Imports

# Python libraries
from __future__ import print_function

# Communication imports
from .exceptions import RobotCommandException
from .objects import PoseObject
from .enums import RobotErrors
from .niryo_ros import NiryoRos


class RobotCommander(object):
    def __init__(self, client):
        assert (isinstance(client, NiryoRos))

        self._client = client

        self._services = None
        self._topics = None
        self._actions = None

    def __str__(self):
        return "NiryoRobot"

    def __repr__(self):
        return self.__str__()

    @property
    def client(self):
        return self._client

    # Parameters checker
    def _check_enum_belonging(self, value, enum_):
        """
        Check if a value belong to an enum
        """
        if value not in enum_:
            self._raise_exception_expected_choice([v for v in enum_], value)

    def _check_list_belonging(self, value, list_):
        """
        Check if a value belong to a list
        """
        if value not in list_:
            self._raise_exception_expected_choice(list_, value)

    def _check_range_belonging(self, value, range_min, range_max):
        """
        Check if a value belong to a range
        """
        if not range_min <= value <= range_max:
            self._raise_exception_expected_range(range_min, range_max, value)

    def _check_dict_belonging(self, value, dict_):
        """
        Check if a value belong to a dictionary
        """
        if value not in dict_.keys():
            self._raise_exception_expected_choice(dict_.keys(), value)

    def _check_type(self, value, type_):
        if type(value) is not type_:
            self._raise_exception_expected_type(type_.__name__, value)

    def _check_instance(self, value, type_):
        if not isinstance(value, type_):
            self._raise_exception_expected_type(
                type_.__name__ if not isinstance(type_, tuple) else " or ".join([type__.__name__ for type__ in type_]),
                value)

    def _check_not_instance(self, value, type_):
        if isinstance(value, type_):
            self._raise_exception_unexpected_type(
                type_.__name__ if not isinstance(type_, tuple) else " or ".join([type__.__name__ for type__ in type_]),
                value)

    def _check_list_type(self, values_list, type_):
        for value in values_list:
            self._check_type(value, type_)

    def _check_length(self, values, length):
        if len(values) != length:
            self._raise_exception_length(len(values), length)

    @staticmethod
    def _check_result_status(result):
        if "status" in result:
            if result["status"] < RobotErrors.SUCCESS.value:
                raise RobotCommandException("Error Code : {}\nMessage : {}".format(result["status"], result["message"]))
        elif "success" in result:
            if not result["success"]:
                raise RobotCommandException(
                    "Success : Failure\nMessage : {}".format(result["status"], result["message"]))

    def _map_list(self, list_, type_):
        """
        Try to map a list to another type (Very useful for list like joints
        which are acquired as string)
        """
        try:
            map_list = list(map(type_, list_))
            return map_list
        except ValueError:
            self._raise_exception_expected_type(type_.__name__, list_)

    def _transform_to_type(self, value, type_):
        """
        Try to change value type to another
        """
        try:
            value = type_(value)
            return value
        except ValueError:
            self._raise_exception_expected_type(type_.__name__, value)

    # -- Useful functions
    def _args_pose_to_list(self, *args):
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, PoseObject):
                return arg.to_list()
            else:
                pose_list = arg
        else:
            pose_list = args

        pose_list_float = self._map_list(pose_list, float)
        if len(pose_list_float) != 6:
            self._raise_exception("A pose should contain 6 elements (x, y, z, roll, pitch, yaw)")
        return pose_list_float

    def _args_pose_to_pose_object(self, *args):
        if len(args) == 1:
            arg = args[0]
        else:
            arg = args

        if isinstance(arg, PoseObject):
            return arg
        elif len(arg) == 6:
            pose_list_float = self._map_list(arg, float)
            return PoseObject(*pose_list_float)
        elif len(arg) == 7:
            pose_list_float = self._map_list(arg, float)
            pose_list_rpy = pose_list_float[:3] + list(PoseObject.quaternion_to_euler_angle(*pose_list_float[3:]))
            return PoseObject(*pose_list_rpy)
        else:
            self._raise_exception("A pose should contain either 6 elements (x, y, z, roll, pitch, yaw), "
                                  "either 7 elements (x, y, z, qx, qy, qz, qw), either a PoseObject.")

    def _args_joints_to_list(self, *args):
        """
        Convert args into a list
        Either if args = (1.1,5.6,-6.7) or args = ([1.1,5.6,-6.7],) , the
        function will return (1.1,5.6,-6.7)

        :param args: Union[list, tuple]
        :return: list of float
        """
        if len(args) == 1:
            args = args[0]

        joints = self._map_list(args, float)
        if len(joints) != 6:
            self._raise_exception("The robot has 6 joints")

        return joints

    # Error Handlers
    def _raise_exception_expected_choice(self, expected_choice, given):
        raise RobotCommandException("Expected one of the following: {}.\nGiven: {}".format(expected_choice, given))

    def _raise_exception_expected_type(self, expected_type, given):
        raise RobotCommandException("Expected type: {}.\nGiven: {}".format(expected_type, given))

    def _raise_exception_unexpected_type(self, unexpected_type, given):
        raise RobotCommandException("Unexpected type: {}.\nGiven: {}".format(unexpected_type, given))

    def _raise_exception_expected_range(self, range_min, range_max, given):
        raise RobotCommandException(
            "Expected the following condition: {} <= value <= {}\nGiven: {}".format(range_min, range_max, given))

    def _raise_exception_length(self, expected_length, given):
        raise RobotCommandException("Expected length: {}.\n Given: {}".format(expected_length, given))

    def _raise_exception(self, message):
        raise RobotCommandException("Exception message : {}".format(message))



from pyniryo2.niryo_topic import NiryoTopic
from pyniryo2.led_ring.objects import LedRingStatusObject


class LedRingTopics(object):

    def __init__(self, client):
        self.__client = client

        self.led_ring_status_topic = NiryoTopic(self.__client,
                                                '/niryo_robot_led_ring/led_ring_status',
                                                'niryo_robot_led_ring/LedRingStatus',
                                                led_status_topic_conversion)


def led_status_topic_conversion(msg):
    led_status = LedRingStatusObject()
    led_status.init_from_message(msg)
    return led_status



"""
Little Script which shows how to read a position with PyNiryo2
"""
# !/usr/bin/env python

from pyniryo2 import NiryoRobot
import sys

if sys.version_info[0] == 2:
    input_func = raw_input
else:
    input_func = input

simulation = "-simu" in sys.argv
robot_ip_address_rpi = "10.10.10.10"
robot_ip_address_simu = "127.0.0.1"
robot_ip_address = robot_ip_address_simu if simulation else robot_ip_address_rpi

# Connecting to robot
niryo_robot = NiryoRobot(ip_address=robot_ip_address)

niryo_robot.arm.calibrate_auto()

while "User to not quit":
    # Going to learning mode
    niryo_robot.arm.set_learning_mode(True)

    input_func("Press enter to switch to turn learning mode off")
    niryo_robot.arm.set_learning_mode(False)

    # Reading pose and displaying it
    pose = niryo_robot.arm.get_pose()
    print(pose)
    if input_func("Press enter to get new pose / q to leave ") == 'q':
        break

niryo_robot.arm.set_learning_mode(True)
# Close
niryo_robot.end()




# !/usr/bin/env python

# Imports
from pyniryo2 import *
from threading import Event

robot_ip = "xxx.xxx.xxx.xxx"
robot_ip_address_local = "127.0.0.1"

# Events
update_tool_event = Event()
update_tool_event.clear()

calibrated_event = Event()
calibrated_event.clear()

# Poses
pose_1 = PoseObject()

pose_2 = PoseObject()

# Callbacks
def update_tool_success_callback(result):
    update_tool_event.set()
    print ('Update Tool: ', result['message'])

def update_tool_error_callback(result):
    print ('Update Tool: ', result['message'])

def calibrate_success_callback(result):
    calibrated_event.set()
    print ('Calibrate Callback: ', result["message"])

def calibrate_error_callback(result):
    print ('Calibrate Callback: ', result["message"])

    """
    Add Callbacks Here
    """

def action_function(robot):

    """
    Don't put niryo_robot in parameter, it will take the pyniryo2 package
    add your function here
    """


if __name__ == "__main__":

    # Connect to robot
    robot = NiryoRobot(robot_ip)

    # Calibrate robot if robot needs calibration
    robot.arm.calibrate_auto(callback=calibrate_success_callback, errback=calibrate_error_callback)
    calibrated_event.wait(20)
    if not calibrated_event.is_set():
        quit

    robot.tool.update_tool(callback=update_tool_success_callback, errback=update_tool_error_callback)
    update_tool_event.wait()

    action_function(robot)

    # Releasing connection
    robot.arm.go_to_sleep()
    robot.end()



import roslibpy

from pyniryo2.utils import pose_list_to_dict, point_list_to_dict

from .enums import ManageFrames
from .objects import DynamicFrameInfo

class FramesServices(object):

    def __init__(self, client):
        self.__client = client

        self.get_frame_from_name_service = roslibpy.Service(self.__client,
                                                            '/niryo_robot_poses_handlers/get_dynamic_frame',
                                                            'niryo_robot_poses_handlers/GetDynamicFrame')

        self.manage_frame_service = roslibpy.Service(self.__client,
                                                     '/niryo_robot_poses_handlers/manage_dynamic_frame',
                                                     'niryo_robot_poses_handlers/ManageDynamicFrame')

        self.get_frame_list_service = roslibpy.Service(self.__client,
                                                       '/niryo_robot_poses_handlers/get_dynamic_frame_list',
                                                       'niryo_robot_msgs/GetNameDescriptionList')

    @staticmethod
    def get_dynamic_frame_from_name_request(frame_name):
        return roslibpy.ServiceRequest({"name": frame_name})

    @staticmethod
    def get_dynamic_frame_from_name_response_to_named_tuple(response):
        response = response["dynamic_frame"]
        frame = DynamicFrameInfo(response["name"], response["description"], 
                                 [response["position"]["x"], response["position"]["y"], response["position"]["z"]],
                                 [response["rpy"]["roll"], response["rpy"]["pitch"], response["rpy"]["yaw"]])
        return frame

    @staticmethod
    def save_dynamic_frame_from_poses_request(frame_name, description, poses_list, belong_to_workspace):
        dynamic_frame = {"name": frame_name, "description": description,
                         "poses": [pose_list_to_dict(pose) for pose in poses_list],
                         "belong_to_workspace": belong_to_workspace}
        return roslibpy.ServiceRequest(
            {"cmd": ManageFrames.SAVE.value, "dynamic_frame": dynamic_frame})

    @staticmethod
    def save_dynamic_frame_from_points_request(frame_name, description, points_list, belong_to_workspace):
        dynamic_frame = {"name": frame_name, "description": description, "points": points_list,
                         "belong_to_workspace": belong_to_workspace}
        return roslibpy.ServiceRequest(
            {"cmd": ManageFrames.SAVE_WITH_POINTS.value, "dynamic_frame": dynamic_frame})

    @staticmethod
    def edit_dynamic_frame_request(frame_name, new_frame_name, new_description):
        dynamic_frame = {"name": frame_name, "new_name": new_frame_name, "description": new_description}
        return roslibpy.ServiceRequest({"cmd": ManageFrames.EDIT.value, "dynamic_frame": dynamic_frame})

    @staticmethod
    def delete_dynamic_frame_request(frame_name, belong_to_workspace):
        return roslibpy.ServiceRequest({"cmd": ManageFrames.DELETE.value,
                                        "dynamic_frame": {"name": frame_name,
                                                          "belong_to_workspace": belong_to_workspace}})

    @staticmethod
    def get_saved_dynamic_frame_list_request():
        return roslibpy.ServiceRequest()

    @staticmethod
    def get_saved_dynamic_frame_list_response_to_list(response):
        list_name = response["name_list"]
        desc_list = response["description_list"]
        dict={}
        for x, y in zip(list_name, desc_list):
            dict[x] = y       
        return dict



from enum import Enum, unique


@unique
class PinMode(Enum):
    """
    Enumeration of Pin Modes
    """
    OUTPUT = 0
    INPUT = 1


@unique
class PinState(Enum):
    """
    Pin State is either LOW or HIGH
    """
    LOW = False
    HIGH = True


@unique
class PinID(Enum):
    """
    Enumeration of Robot Pins
    """
    GPIO_1A = "1A"
    GPIO_1B = "1B"
    GPIO_1C = "1C"
    GPIO_2A = "2A"
    GPIO_2B = "2B"
    GPIO_2C = "2C"

    SW_1 = "SW1"
    SW_2 = "SW2"

    DO1 = "DO1"
    DO2 = "DO2"
    DO3 = "DO3"
    DO4 = "DO4"
    DI1 = "DI1"
    DI2 = "DI2"
    DI3 = "DI3"
    DI4 = "DI4"
    DI5 = "DI5"

    AI1 = "AI1"
    AI2 = "AI2"
    AO1 = "AO1"
    AO2 = "AO2"



#!/usr/bin/env python
import time
import unittest
import numpy as np
from threading import Event

from pyniryo2.exceptions import RobotCommandException
from pyniryo2.niryo_topic import NiryoTopic
from pyniryo2.enums import RobotErrors
from pyniryo2.objects import PoseObject
from pyniryo2.niryo_ros import NiryoRos

from pyniryo2.tool.tool import Tool
from pyniryo2.tool.enums import ToolID

from pyniryo2.arm.arm import Arm

from pyniryo2.io.io import IO
from pyniryo2.io.enums import PinID, PinMode

robot_ip_address = "127.0.0.1"
port = 9090

test_order = ["test_tool_id",
              "test_update_tool",
              "test_electromagnet",
              "test_grippers",
              "test_vacuum_pump",
              "test_tcp"
              ]


class BaseTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.client = NiryoRos(ip_address=robot_ip_address, port=port)
        cls.tool = Tool(cls.client)
        cls.io = IO(cls.client)
        cls.arm = Arm(cls.client)

    @classmethod
    def tearDownClass(cls):
        cls.client.close()

    @staticmethod
    def assertAlmostEqualVector(a, b, decimal=1):
        np.testing.assert_almost_equal(a, b, decimal)


# noinspection PyTypeChecker
class TestTool(BaseTest):

    def test_tool_id(self):
        self.assertIsNone(time.sleep(1))
        self.assertIsInstance(self.tool.tool, ToolID)
        self.assertIsInstance(self.tool.get_current_tool_id, NiryoTopic)
        self.assertIsInstance(self.tool.get_current_tool_id(), ToolID)

        tool_id_event = Event()
        tool_id_event.clear()

        def tool_id_callback(tool_id):
            tool_id_event.set()
            self.assertIsInstance(tool_id, ToolID)

        self.assertIsNone(self.tool.get_current_tool_id.subscribe(tool_id_callback))
        self.assertTrue(tool_id_event.wait(10))
        self.assertIsNone(self.tool.get_current_tool_id.unsubscribe())

    def test_update_tool(self):
        self.assertIsNone(time.sleep(1))
        self.assertIsNone(self.tool.update_tool())
        self.assertIsNone(time.sleep(1))

        tool_update_event = Event()
        tool_update_event.clear()

        def tool_update_callback(result):
            self.assertTrue(result["status"] >= RobotErrors.SUCCESS.value)
            tool_update_event.set()

        self.assertIsNone(self.tool.update_tool(tool_update_callback))
        self.assertTrue(tool_update_event.wait(10))

    def test_electromagnet(self):
        # Setup
        with self.assertRaises(RobotCommandException):
            self.tool.setup_electromagnet(0)

        pin = PinID.DO2 if self.tool.client.hardware_version == 'ned2' else PinID.GPIO_1A
        if self.tool.client.hardware_version != 'ned2':
            self.assertIsNone(self.io.set_pin_mode(pin, PinMode.INPUT))
        self.assertIsNone(self.tool.setup_electromagnet(pin))
        self.assertEqual(self.tool.get_current_tool_id(), ToolID.ELECTROMAGNET_1)
        self.assertEqual(self.io.get_digital_io_state(pin).mode, PinMode.OUTPUT)

        self.assertIsNone(self.tool.activate_electromagnet())
        time.sleep(0.5)
        self.assertEqual(self.io.digital_read(pin), True)
        self.assertIsNone(self.tool.deactivate_electromagnet())
        time.sleep(0.5)
        self.assertEqual(self.io.digital_read(pin), False)

        tool_event = Event()
        tool_event.clear()

        def tool_callback(_):
            tool_event.set()

        self.assertIsNone(self.tool.activate_electromagnet(callback=tool_callback))
        self.assertTrue(tool_event.wait(10))
        self.assertEqual(self.io.digital_read(pin), True)
        tool_event.clear()

        self.assertIsNone(self.tool.deactivate_electromagnet(callback=tool_callback))
        self.assertTrue(tool_event.wait(10))
        self.assertEqual(self.io.digital_read(pin), False)

        self.assertIsNone(self.tool.grasp_with_tool())
        self.assertEqual(self.io.digital_read(pin), True)
        self.assertIsNone(self.tool.release_with_tool())
        self.assertEqual(self.io.digital_read(pin), False)

        # Exceptions
        with self.assertRaises(RobotCommandException):
            self.tool.activate_electromagnet(pin_id=5)
        with self.assertRaises(RobotCommandException):
            self.tool.deactivate_electromagnet(pin_id=5)
        self.assertIsNone(self.tool.update_tool())
        with self.assertRaises(RobotCommandException):
            self.tool.activate_electromagnet()
        with self.assertRaises(RobotCommandException):
            self.tool.deactivate_electromagnet()

    def test_grippers(self):
        self.assertIsNone(self.tool.update_tool())

        tool_event = Event()

        def tool_callback(_):
            tool_event.set()

        if self.tool.get_current_tool_id() in [ToolID.GRIPPER_1, ToolID.GRIPPER_2, ToolID.GRIPPER_3]:
            self.assertIsNone(self.tool.open_gripper())
            self.assertIsNone(self.tool.open_gripper(800))
            self.assertIsNone(self.tool.open_gripper(speed=800.5))
            self.assertIsNone(self.tool.open_gripper(max_torque_percentage=100))
            self.assertIsNone(self.tool.open_gripper(hold_torque_percentage=50))
            self.assertIsNone(self.tool.open_gripper(max_torque_percentage=100, hold_torque_percentage=50))
            tool_event.clear()
            self.assertIsNone(self.tool.open_gripper(callback=tool_callback))
            self.assertTrue(tool_event.wait(10))

            with self.assertRaises(RobotCommandException):
                self.tool.open_gripper(-10)
            with self.assertRaises(RobotCommandException):
                self.tool.open_gripper(0)
            with self.assertRaises(RobotCommandException):
                self.tool.open_gripper(1001)
            with self.assertRaises(RobotCommandException):
                self.tool.open_gripper(max_torque_percentage=200)
            with self.assertRaises(RobotCommandException):
                self.tool.open_gripper(max_torque_percentage=-10)
            with self.assertRaises(RobotCommandException):
                self.tool.open_gripper(hold_torque_percentage=140)
            with self.assertRaises(RobotCommandException):
                self.tool.open_gripper(hold_torque_percentage=-20)

            self.assertIsNone(self.tool.close_gripper())
            self.assertIsNone(self.tool.close_gripper(800))
            self.assertIsNone(self.tool.close_gripper(speed=800.5))
            self.assertIsNone(self.tool.close_gripper(max_torque_percentage=100))
            self.assertIsNone(self.tool.close_gripper(hold_torque_percentage=50))
            self.assertIsNone(self.tool.close_gripper(max_torque_percentage=100, hold_torque_percentage=50))
            tool_event.clear()
            self.assertIsNone(self.tool.close_gripper(callback=tool_callback))
            self.assertTrue(tool_event.wait(10))

            with self.assertRaises(RobotCommandException):
                self.tool.close_gripper(-10)
            with self.assertRaises(RobotCommandException):
                self.tool.close_gripper(0)
            with self.assertRaises(RobotCommandException):
                self.tool.close_gripper(1001)
            with self.assertRaises(RobotCommandException):
                self.tool.close_gripper(max_torque_percentage=200)
            with self.assertRaises(RobotCommandException):
                self.tool.close_gripper(max_torque_percentage=-10)
            with self.assertRaises(RobotCommandException):
                self.tool.close_gripper(hold_torque_percentage=140)
            with self.assertRaises(RobotCommandException):
                self.tool.close_gripper(hold_torque_percentage=-20)

            self.assertIsNone(self.tool.grasp_with_tool())
            self.assertIsNone(self.tool.release_with_tool())

            tool_event.clear()
            self.assertIsNone(self.tool.grasp_with_tool(callback=tool_callback))
            self.assertTrue(tool_event.wait(10))

            tool_event.clear()
            self.assertIsNone(self.tool.release_with_tool(callback=tool_callback))
            self.assertTrue(tool_event.wait(10))

        # Exceptions
        pin = PinID.DO4 if self.tool.client.hardware_version == 'ned2' else PinID.GPIO_1A
        self.assertIsNone(self.tool.setup_electromagnet(pin))
        with self.assertRaises(RobotCommandException):
            self.tool.open_gripper()
        with self.assertRaises(RobotCommandException):
            self.tool.close_gripper()

    def test_vacuum_pump(self):
        self.assertIsNone(self.tool.update_tool())

        tool_event = Event()

        def tool_callback(_):
            tool_event.set()

        if self.tool.get_current_tool_id() in [ToolID.VACUUM_PUMP_1, ]:
            self.assertIsNone(self.tool.pull_air_vacuum_pump())
            tool_event.clear()
            self.assertIsNone(self.tool.pull_air_vacuum_pump(callback=tool_callback))
            self.assertTrue(tool_event.wait(10))

            self.assertIsNone(self.tool.push_air_vacuum_pump())
            tool_event.clear()
            self.assertIsNone(self.tool.push_air_vacuum_pump(callback=tool_callback))
            self.assertTrue(tool_event.wait(10))

            self.assertIsNone(self.tool.grasp_with_tool())
            self.assertIsNone(self.tool.release_with_tool())

            tool_event.clear()
            self.assertIsNone(self.tool.grasp_with_tool(callback=tool_callback))
            self.assertTrue(tool_event.wait(10))

            tool_event.clear()
            self.assertIsNone(self.tool.release_with_tool(callback=tool_callback))
            self.assertTrue(tool_event.wait(10))

        # Exceptions
        pin = PinID.DO4 if self.tool.client.hardware_version == 'ned2' else PinID.GPIO_1A
        self.assertIsNone(self.tool.setup_electromagnet(pin))
        with self.assertRaises(RobotCommandException):
            self.tool.pull_air_vacuum_pump()
        with self.assertRaises(RobotCommandException):
            self.tool.push_air_vacuum_pump()

    def test_tcp(self):
        self.assertIsNone(time.sleep(1))
        self.assertIsNone(self.tool.reset_tcp())
        self.assertIsNone(self.tool.enable_tcp(False))

        self.assertIsNone(time.sleep(2))
        test_pose = [0.3, 0.0, 0.2, 0.0, 1.57, 0.0]
        self.assertIsNone(self.arm.move_pose(test_pose))
        self.assertAlmostEqualVector(self.arm.pose.to_list()[:3], test_pose[:3])

        # Test Set TCP
        self.assertIsNone(self.tool.set_tcp(0.01, 0.02, 0.03, 0.0, 0.0, 0.0))
        new_pose = test_pose[:3]
        new_pose[2] -= 0.01
        new_pose[1] += 0.02
        new_pose[0] += 0.03
        self.assertAlmostEqualVector(self.arm.pose.to_list()[:3], new_pose, 2)

        self.assertIsNone(self.tool.set_tcp(PoseObject(0.1, 0.0, 0.0, 0.0, 0.0, 0.0)))
        new_pose = test_pose[:]
        new_pose[2] -= 0.1
        self.assertAlmostEqualVector(self.arm.pose.to_list()[:3], new_pose[:3])

        # Enable/Disable
        self.assertIsNone(self.tool.enable_tcp(False))
        self.assertAlmostEqualVector(self.arm.pose.to_list()[:3], test_pose[:3])
        self.assertIsNone(self.tool.enable_tcp())
        self.assertAlmostEqualVector(self.arm.pose.to_list()[:3], new_pose[:3])

        # Reset
        self.assertIsNone(self.tool.reset_tcp())
        self.assertAlmostEqualVector(self.arm.pose.to_list()[:3], test_pose[:3])

        # Exceptions
        with self.assertRaises(RobotCommandException):
            self.tool.set_tcp(5 * [0.0])

        with self.assertRaises(RobotCommandException):
            self.tool.set_tcp(7 * [0.0])

        with self.assertRaises(RobotCommandException):
            self.tool.set_tcp(0.0, 0.0)

        with self.assertRaises(RobotCommandException):
            self.tool.set_tcp(*(8 * [0.0]))

        with self.assertRaises(RobotCommandException):
            self.tool.set_tcp("a")


def suite():
    suite = unittest.TestSuite()
    for function_name in test_order:
        suite.addTest(TestTool(function_name))
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



# - Imports
from __future__ import print_function
from numpy import isin

# Python libraries
import roslibpy
import time

# Communication imports
from pyniryo2.robot_commander import RobotCommander
from pyniryo2.enums import RobotErrors
from pyniryo2.objects import PoseObject
from pyniryo2.utils import pose_dict_to_list
from pyniryo2.exceptions import RobotCommandException

from .enums import CalibrateMode, RobotAxis, JogShift
from .services import ArmServices
from .topics import ArmTopics
from .actions import ArmActions


class Arm(RobotCommander):
    # --- Public functions --- #
    def __init__(self, client):
        """
        Arm robot functions

        Example: ::

            ros_instance = NiryoRos("10.10.10.10") # Hotspot
            arm_interface = Arm(ros_instance)

        :param client: Niryo ROS client
        :type client: NiryoRos
        """
        super(Arm, self).__init__(client)

        self._services = ArmServices(self._client)
        self._topics = ArmTopics(self._client)
        self._actions = ArmActions(self._client)

        self.__action_timeout = 10

    # - Main purpose
    # -- Calibration

    def calibrate(self, calibrate_mode, callback=None, errback=None, timeout=None):
        """
        Calibrates (manually or automatically) motors. Automatic calibration will do nothing
        if motors are already calibrated

        Examples: ::

            # Synchronous use
            arm.calibrate(CalibrateMode.MANUAL)
            arm.calibrate(CalibrateMode.AUTO)

            # Asynchronous use
            def calibration_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Calibration failed")
                else:
                    print("Calibration completed with success")

            arm.calibrate(CalibrateMode.AUTO, calibration_callback)


        :param calibrate_mode: Auto or Manual
        :type calibrate_mode: CalibrateMode
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param errback: Callback invoked on error.
        :type errback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """

        def wait_calibration_end():
            for _ in range(40):
                time.sleep(0.05)
                hardware_status = self._topics.hardware_status_topic()
                if hardware_status and not (
                        hardware_status.calibration_in_progress or hardware_status.calibration_needed):
                    return
            raise RobotCommandException("Calibration timeout")

        self._check_enum_belonging(calibrate_mode, CalibrateMode)
        request = roslibpy.ServiceRequest()
        request["value"] = calibrate_mode.value

        if callback is not None:
            self._services.request_calibration_service.call(request, callback=callback, errback=errback,
                                                            timeout=timeout)
            return
        else:
            resp = self._services.request_calibration_service.call(request, callback=None, errback=errback,
                                                                   timeout=timeout)
            self._check_result_status(resp)
            return wait_calibration_end()

    def calibrate_auto(self, callback=None, errback=None, timeout=None):
        """
        Starts a automatic motors calibration if motors are not calibrated yet.

        Examples: ::

            # Synchronous use
            arm.calibrate_auto()

            # Asynchronous use
            def calibration_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Calibration failed")
                else:
                    print("Calibration completed with success")

            arm.calibrate_auto(calibration_callback)

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param errback: Callback invoked on error.
        :type errback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        return self.calibrate(CalibrateMode.AUTO, callback, errback, timeout)

    def request_new_calibration(self, callback=None, errback=None, timeout=None):
        """
        Starts a automatic motors calibration even if motors are calibrated yet.

        Examples: ::

            # Synchronous use
            arm.request_new_calibration()

            # Asynchronous use
            def calibration_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Calibration failed")
                else:
                    print("Calibration completed with success")

            arm.request_new_calibration(calibration_callback)

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param errback: Callback invoked on error.
        :type errback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        self.reset_calibration()
        return self.calibrate_auto(callback, errback, timeout)

    def reset_calibration(self, timeout=2):
        """
        Resets current calibration status. A new calibration is then necessary.

        :param timeout: Timeout for the operation, in seconds.
        :type timeout: float
        :rtype: None
        """
        request = self._services.get_trigger_request()
        self._services.request_new_calibration_service.call(request)

        start_time = time.time()
        while time.time() - start_time < timeout:
            hardware_status = self._topics.hardware_status_topic()
            if hardware_status and hardware_status.calibration_needed:
                return
            time.sleep(0.05)

    def need_calibration(self):
        """
        Returns a bool indicating whereas the robot motors need to be calibrate.

        :return: True if calibration is needed, False otherwise.
        :rtype: bool
        """
        hardware_status = self._topics.hardware_status_topic()
        return hardware_status.calibration_needed

    # - Hardware Status

    @property
    def hardware_status(self):
        """
        Returns the hardware state client which can be used synchronously or asynchronously
        to obtain the hardware state.

        Examples: ::

            # Get last value
            arm.hardware_status()
            arm.hardware_status.value

            # Subscribe a callback
            def hs_callback(msg):
                print msg.voltage

            arm.hardware_status.subscribe(hs_callback)
            arm.hardware_status.unsubscribe()

        :return: hardware state topic instance
        :rtype: NiryoTopic
        """
        return self._topics.hardware_status_topic

    # -- Learning mode

    @property
    def learning_mode(self):
        """
        Returns the learning mode client which can be used synchronously or asynchronously
        to obtain the learning mode state.
        The learning mode client returns a boolean value.

        Examples: ::

            # Get last value
            arm.learning_mode()
            if arm.learning_mode.value:
                print("Learning mode enabled"))

            # Subscribe a callback
            def lm_callback(is_learning_mode_enabled):
                print is_learning_mode_enabled

            arm.learning_mode.subscribe(lm_callback)
            arm.learning_mode.unsubscribe()

        :return: learning mode state topic instance
        :rtype: NiryoTopic
        """
        return self._topics.learning_mode_state_topic

    def get_learning_mode(self):
        """
        Get learning mode state.

        :return: ``True`` if learning mode is on
        :rtype: bool
        """
        return self._topics.learning_mode_state_topic()

    @learning_mode.setter
    def learning_mode(self, enabled):
        """
        Set learning mode if param is ``True``, else turn it off

        :param enabled: ``True`` or ``False``
        :type enabled: bool
        """
        self.set_learning_mode(enabled)

    def set_learning_mode(self, enabled):
        """
        Set learning mode if param is ``True``, else turn it off

        :param enabled: ``True`` or ``False``
        :type enabled: bool
        :rtype: None
        """
        self._check_type(enabled, bool)
        req = self._services.get_learning_mode_request(enabled)
        resp = self._services.activate_learning_mode_service.call(req)
        self._check_result_status(resp)

    # - Get Joints/Pose

    @property
    def joints_state(self):
        """
        Get the joints state topic which can be used synchronously or asynchronously to obtain the joints state.
        The joints state topic returns a JointStateObject.

        It can be used as follows:: ::

            # Get last joint state
            joint_state = arm.joints_state()
            joint_state = arm.joints_state.value

            joint_names = arm.joints_state().name
            joint_positions = arm.joints_state().position
            joint_velocities = arm.joints_state.value.velocity

            # Raise a callback at each new value
            from __future__ import print_function

            arm.joints_state.subscribe(lambda message: print(message.position))
            arm.joints_state.unsubscribe()


        :return: Joint states topic.
        :rtype: NiryoTopic
        """
        return self._topics.joint_states_topic

    @property
    def joints(self):
        """
        Get joints value in radians

        :return: List of joints value
        :rtype: list[float]
        """
        return self.get_joints()

    def get_joints(self):
        """
        Get joints value in radians
        You can also use a getter ::

            joints = arm.get_joints()
            joints = arm.joints

        :return: List of joints value
        :rtype: list[float]
        """
        return self._topics.joint_states_topic().position[:6]

    @property
    def pose(self):
        """
        Get end effector link pose as [x, y, z, roll, pitch, yaw].
        x, y & z are expressed in meters / roll, pitch & yaw are expressed in radians

        You can also use a getter ::

            pose = arm.pose
            pose_list = arm.pose.to_list()
            x, y, z, roll, pitch, yaw = arm.pose.to_list()

        :return: end effector link pose
        :rtype: PoseObject
        """
        return self._topics.robot_state_topic()

    @property
    def get_pose(self):
        """
        Get the end effector link pose topic which can be used synchronously or asynchronously
        to obtain the end effector link pose.
        The joints state topic returns a PoseObject.
        x, y & z are expressed in meters / roll, pitch & yaw are expressed in radians

        See below some usage ::

            pose = arm.get_pose()
            pose = arm.get_pose.value
            pose_list = arm.get_pose().to_list()
            x, y, z, roll, pitch, yaw = arm.get_pose().to_list()

            arm.get_pose.subscribe(callback)
            arm.get_pose.unsubscribe()

        :return: end effector link pose topic
        :rtype: NiryoTopic
        """
        return self._topics.robot_state_topic

    def get_pose_quat(self):
        """
        Get end effector link pose in Quaternion coordinates

        :return: Position and quaternion coordinates concatenated in a list : [x, y, z, qx, qy, qz, qw]
        :rtype: list[float]
        """
        return self._topics.robot_state_topic().quaternion

    # --  Simple Move

    @joints.setter
    def joints(self, joints):
        """
        Move robot joints. Joints are expressed in radians.

        :param joints: a list of 6 joints
        :type joints: Union[list[float], tuple[float]]
        """
        self.move_joints(joints)

    def move_joints(self, joints, callback=None):
        """
        Move robot joints. Joints are expressed in radians.
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        All lines of the next example realize the same operation: ::

            arm.joints = [0.2, 0.1, 0.3, 0.0, 0.5, 0.0]
            arm.move_joints([0.2, 0.1, 0.3, 0.0, 0.5, 0.0])

            def move_callback(_):
                print("Move completed")

            arm.move_joints([0chronous use
            arm.calibrate(CalibrateMode.MANUAL)
            arm.calibrate(CalibrateMode.AUTO)

            # Asynchronous use
            def calibration_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Calibration failed")
                else:
                    print("Calibration completed with success")

            arm.calibrate(CalibrateMode.AUTO, calibration_callback)

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param joints: a list of 6 joints
        :type joints: Union[list[float], tuple[float]]
        :rtype: None
        """
        joint_list = self._args_joints_to_list(joints)
        goal = self._actions.get_move_joints_goal(joint_list)
        goal.send(result_callback=callback)
        if callback is None:
            _result = goal.wait(self.__action_timeout)

    @pose.setter
    def pose(self, pose):
        """
        Move robot end effector pose to a (x, y, z, roll, pitch, yaw) pose.
        x, y & z are expressed in meters / roll, pitch & yaw are expressed in radians

        :param pose: either a list of 6 coordinates or a ``PoseObject``
        :type pose: Union[tuple[float], list[float], PoseObject]
        """
        self.move_pose(pose)

    def _get_transform_pose(self, source_frame, local_frame, position, rpy):
        """
        Transform a pose from a source_frame to a pose in a local_frame

        :param source_frame: name of the source_frame
        :type source_frame: str
        :param local_frame: name of the local_frame
        :type local_frame: str
        :param position: position of the pose to convert
        :type position: dict[x, y, z]
        :param rpy: orientation of the pose to convert
        :type rpy: dict[roll, pitch, yaw]
        """
        req = self._services.get_transform_pose_request(source_frame, local_frame, position, rpy)
        response = self._services.get_transform_pose_service.call(req)
        return response["position"], response["rpy"]

    def move_pose(self, pose, frame="", callback=None):
        """
        Move robot end effector pose to a (x, y, z, roll, pitch, yaw) pose in the frame (frame_name) if defined.
        x, y & z are expressed in meters / roll, pitch & yaw are expressed in radians
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        All lines of the next example realize the same operation: ::

            arm.pose = [0.2, 0.1, 0.3, 0.0, 0.5, 0.0]
            arm.move_pose([0.2, 0.1, 0.3, 0.0, 0.5, 0.0])
            arm.move_pose(PoseObject(0.2, 0.1, 0.3, 0.0, 0.5, 0.0))
            arm.move_pose([0.2, 0.1, 0.3, 0.0, 0.5, 0.0], "default_frame")
            arm.move_pose(PoseObject(0.2, 0.1, 0.3, 0.0, 0.5, 0.0), "default_frame")

            def move_callback(_):
                print("Move completed")

            arm.move_joints([0.2, 0.1, 0.3, 0.0, 0.5, 0.0], callback=move_callback)

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param pose: either a list of 6 coordinates or a ``PoseObject``
        :type pose: Union[tuple[float], list[float], PoseObject]
        :param frame: name of the frame
        :type param: str

        :rtype: None
        """
        if (frame != ""):
            position = {}
            rpy = {}
            if (isinstance(pose, PoseObject)):
                position = {"x": pose.x, "y": pose.y, "z": pose.z}
                rpy = {"roll": pose.roll, "pitch": pose.pitch, "yaw": pose.yaw}
            else:
                position = {"x": pose[0], "y": pose[1], "z": pose[2]}
                rpy = {"roll": pose[3], "pitch": pose[4], "yaw": pose[5]}
            frame_position, frame_rpy = self._get_transform_pose("world", frame, position, rpy)

            pose_list = [frame_position[axis] for axis in ["x", "y", "z"]] + \
                [frame_rpy[axis] for axis in ["roll", "pitch", "yaw"]]
        else:
            pose_list = self.__args_pose_to_list(pose)

        goal = self._actions.get_move_pose_goal(pose_list)
        goal.send(result_callback=callback)
        if callback is None:
            _result = goal.wait(self.__action_timeout)

    def move_to_home_pose(self, callback=None):
        """
        Move to a position where the forearm lays on shoulder
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :rtype: None
        """
        self.move_joints([0.0, 0.3, -1.3, 0.0, 0.0, 0.0], callback)

    def go_to_sleep(self):
        """
        Go to home pose and activate learning mode

        :rtype: None
        """
        self.move_to_home_pose()
        self.set_learning_mode(True)

    def move_linear_pose(self, pose, frame="", callback=None):
        """
        Move robot end effector pose to a (x, y, z, roll, pitch, yaw) pose in a linear way, in the frame (frame_name) if defined.
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        All lines of the next example realize the same operation: ::

            arm.move_linear_pose([0.2, 0.1, 0.3, 0.0, 0.5, 0.0])
            arm.move_linear_pose(PoseObject(0.2, 0.1, 0.3, 0.0, 0.5, 0.0))
            arm.move_linear_pose([0.2, 0.1, 0.3, 0.0, 0.5, 0.0], "default_frame")
            arm.move_linear_pose(PoseObject(0.2, 0.1, 0.3, 0.0, 0.5, 0.0), "default_frame")

            def move_callback(_):
                print("Move completed")

            arm.move_linear_pose([0.2, 0.1, 0.3, 0.0, 0.5, 0.0], callback=move_callback)

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param pose: either or a list of 6 coordinates or a PoseObject
        :type pose: Union[tuple[float], list[float], PoseObject]
        :rtype: None
        :param frame: name of the frame
        :type param: str
        """
        if (frame != ""):
            position = {}
            rpy = {}
            if (isinstance(pose, PoseObject)):
                position = {"x": pose.x, "y": pose.y, "z": pose.z}
                rpy = {"roll": pose.roll, "pitch": pose.pitch, "yaw": pose.yaw}
            else:
                position = {"x": pose[0], "y": pose[1], "z": pose[2]}
                rpy = {"roll": pose[3], "pitch": pose[4], "yaw": pose[5]}
            frame_position, frame_rpy = self._get_transform_pose("world", frame, position, rpy)

            pose_list = [frame_position[axis] for axis in ["x", "y", "z"]] + \
                [frame_rpy[axis] for axis in ["roll", "pitch", "yaw"]]
        else:
            pose_list = self.__args_pose_to_list(pose)

        goal = self._actions.get_move_linear_pose_goal(pose_list)
        goal.send(result_callback=callback)
        _result = goal.wait(self.__action_timeout)

    def stop_move(self, callback=None, errback=None, timeout=None):
        """
        Stop a current execution of move_pose, move_joint or move_linear_pose.
        The robot will stop at its current position .
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            # Synchronous use
            arm.stop_move()

            # Asynchronous use
            def stop_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Succeeded")
                else:
                    print("Failed")

            arm.stop_move(stop_callback)

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param errback: Callback invoked on error.
        :type errback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        req = self._services.get_trigger_request()
        resp = self._services.stop_arm_service.call(req, callback, errback, timeout)

        self._check_result_status(resp)

    def _calculate_relative_pose(self, frame_name, offset):
        """
        Calculate the pose by a relative movement (x,y,z,roll,pitch,yaw) in the frame (frame_name)

        :param frame_name:, name of the frame
        :type frame_name : str
        :param x:
        :type x: float
        :param y:
        :type y: float
        :param z:
        :type z: float
        :param roll:
        :type roll: float
        :param pitch:
        :type pitch: float
        :param yaw:
        :type yaw: float
        :return: status, message
        :rtype: (int, str)
        """
        import math
        # Get transform
        x, y, z, roll, pitch, yaw = self.get_pose().to_list()

        position = {"x": x, "y": y, "z": z}
        rpy = {"roll": roll, "pitch": pitch, "yaw": yaw}

        frame_position, frame_rpy = self._get_transform_pose(frame_name, "world", position, rpy)

        frame_position["x"] = frame_position["x"] + offset[0]
        frame_position["y"] = frame_position["y"] + offset[1]
        frame_position["z"] = frame_position["z"] + offset[2]

        frame_rpy["roll"] = (frame_rpy["roll"] + offset[3]) % (2*math.pi)
        frame_rpy["pitch"] = (frame_rpy["pitch"] + offset[4]) % (2*math.pi)
        frame_rpy["yaw"] = (frame_rpy["yaw"] + offset[5]) % (2*math.pi)

        world_position, world_rpy = self._get_transform_pose("world", frame_name, frame_position, frame_rpy)

        pose = [world_position["x"], world_position["y"], world_position["z"],
                world_rpy["roll"], world_rpy["pitch"], world_rpy["yaw"]]

        return pose

    def move_relative(self, offset, frame="world"):
        """
        Move robot end of a offset in a frame

        Example: ::

            arm.move_relative([0.05, 0.05, 0.05, 0.3, 0, 0], "default_frame")

        :param offset: list which contains offset of x, y, z, roll, pitch, yaw
        :type offset: list[float]
        :param frame: name of local frame
        :type frame: str
        :return: status, message
        :rtype: (int, str)
        """
        self._check_type(frame, str)
        self._check_type(offset, list)

        pose = self._calculate_relative_pose(frame, offset)

        self.move_pose(pose)

    def move_linear_relative(self, offset, frame="world"):
        """
        Move robot end of a offset by a linear movement in a frame

        Example: ::

            arm.move_linear_relative([0.05, 0.05, 0.05, 0.3, 0, 0], "default_frame")

        :param offset: list which contains offset of x, y, z, roll, pitch, yaw
        :type offset: list[float]
        :param frame: name of local frame
        :type frame: str
        :return: status, message
        :rtype: (int, str)
        """
        self._check_type(frame, str)
        self._check_type(offset, list)
        pose = self._calculate_relative_pose(frame, offset)

        self.move_linear_pose(pose)

    @property
    def get_arm_max_velocity(self):
        """
        Returns the arm max velocity client which can be used synchronously or asynchronously
        to obtain the arm max velocity.
        The arm max velocity client returns an integer value.

        Examples: ::

            # Get last value
            arm.get_arm_max_velocity()
            arm.get_arm_max_velocity.value

            # Subscribe a callback
            def velocity_callback(velocity_percentage):
                print velocity_percentage

            arm.get_arm_max_velocity.subscribe(velocity_callback)
            arm.get_arm_max_velocity.unsubscribe()

        :return: arm velocity topic instance
        :rtype: NiryoTopic
        """
        return self._topics.max_velocity_scaling_factor_topic

    def set_arm_max_velocity(self, percentage_speed):
        """
        Limit arm max velocity to a percentage of its maximum velocity

        :param percentage_speed: Should be between 1 & 100
        :type percentage_speed: int
        :rtype: None
        """
        self._check_range_belonging(percentage_speed, 1, 100)
        req = self._services.get_max_velocity_scaling_factor_request(percentage_speed)
        resp = self._services.set_max_velocity_scaling_factor_service.call(req)
        self._check_result_status(resp)

    # - Shift Pose

    def shift_pose(self, axis, shift_value, callback=None):
        """
        Shift robot end effector pose along one axis
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            self.arm.shift_pose(RobotAxis.X, 0.05)
            self.arm.shift_pose(RobotAxis.Y, -0.05)
            self.arm.shift_pose(RobotAxis.Z, 0.1)
            self.arm.shift_pose(RobotAxis.ROLL, 1.57)
            self.arm.shift_pose(RobotAxis.PITCH, -1.57)
            self.arm.shift_pose(RobotAxis.YAW, 0.78)

            def move_callback(_):
                print("Move completed")

            self.arm.shift_pose(RobotAxis.X, 0.1, move_callback)


        :param axis: Axis along which the robot is shifted
        :type axis: RobotAxis
        :param shift_value: In meter for X/Y/Z and radians for roll/pitch/yaw
        :type shift_value: float
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :rtype: None
        """
        self._check_enum_belonging(axis, RobotAxis)
        shift_value = self._transform_to_type(shift_value, float)

        goal = self._actions.get_shift_pose_goal(axis.value, shift_value)
        goal.send(result_callback=callback)
        if callback is None:
            _result = goal.wait(self.__action_timeout)

    # - Jog

    def set_jog_control(self, enabled):
        """
        Set jog control mode if param is True, else turn it off

        :param enabled: ``True`` or ``False``
        :type enabled: Bool
        :rtype: None
        """
        self._check_type(enabled, bool)
        req = self._services.get_enable_jog_request(enabled)
        resp = self._services.enable_jog_controller_service.call(req)
        self._check_result_status(resp)

    def jog_joints(self, joints_offset, callback=None, errback=None, timeout=None):
        """
        Jog robot joints'.
        Jog corresponds to a shift without motion planning.
        Values are expressed in radians.
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            arm.jog_joints([0.1, 0.0, 0.5, 0.0, 0.0, -1.57])

            def jog_callback(_):
                print("Jog completed")
                arm.set_jog_control(False) # Disable Jog interface

            arm.jog_joints([0.1, 0.0, 0.5, 0.0, 0.0, -1.57], jog_callback)


        :param joints_offset: a list of 6 joints offset
        :type joints_offset: Union[list[float], tuple[float]]
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param errback: Callback invoked on error.
        :type errback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        joints_off = self.__args_joints_to_list(joints_offset)
        req = self._services.get_jog_request(JogShift.JOINTS_SHIFT.value, joints_off)
        self._services.jog_shift_service.call(req, callback, errback, timeout)

    def jog_pose(self, pose_offset, callback=None, errback=None, timeout=None):
        """
        Jog robot end effector pose
        Jog corresponds to a shift without motion planning
        Arguments are [dx, dy, dz, d_roll, d_pitch, d_yaw]
        dx, dy & dz are expressed in meters / d_roll, d_pitch & d_yaw are expressed in radians
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            arm.jog_pose([0.01, 0.0, 0.05, 0.0, 0.0, -1.57])

            def jog_callback(_):
                print("Jog completed")
                arm.set_jog_control(False) # Disable Jog interface

            arm.jog_pose([0.1, 0.0, 0.5, 0.0, 0.0, -1.57], jog_callback)


        :param pose_offset: a list of 6 offset
        :type pose_offset: Union[list[float], tuple[float]]
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param errback: Callback invoked on error.
        :type errback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        pose_offset_list = self.__args_pose_to_list(pose_offset)
        req = self._services.get_jog_request(JogShift.POSE_SHIFT.value, pose_offset_list)
        self._services.jog_shift_service.call(req, callback, errback, timeout)

    # -- Kinematics

    def forward_kinematics(self, *args):
        """
        Compute forward kinematics of a given joints configuration and give the
        associated spatial pose

        Examples: ::

            pose_obj = arm.forward_kinematics(1.57, 0.0, 0.0, 0.78, 0.0, -1.57)
            pose_obj = arm.forward_kinematics([1.57, 0.0, 0.0, 0.78, 0.0, -1.57])

        :param args: either 6 args (1 for each joints) or a list of 6 joints
        :type args: Union[list[float], tuple[float]]
        :rtype: PoseObject
        """
        joints = self.__args_joints_to_list(*args)
        request = self._services.get_forward_kinematics_request(joints)
        response = self._services.forward_kinematics_service.call(request)

        pose_array = pose_dict_to_list(response["pose"])
        pose_object = PoseObject(*pose_array)
        return pose_object

    def inverse_kinematics(self, *args):
        """
        Compute inverse kinematics

        Examples: ::

            joint_list = arm.inverse_kinematics(0.2, 0.0, 0.3, 0.0, 1.57, 0.0)
            joint_list = arm.inverse_kinematics([0.2, 0.0, 0.3, 0.0, 1.57, 0.0])
            joint_list = arm.inverse_kinematics(PoseObject(0.2, 0.0, 0.3, 0.0, 1.57, 0.0))

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a ``PoseObject``
        :type args: Union[tuple[float], list[float], PoseObject]
        :return: List of joints value
        :rtype: list[float]
        """
        pose_list = self.__args_pose_to_list(*args)
        request = self._services.get_inverse_kinematics_request(pose_list)
        response = self._services.inverse_kinematics_service.call(request)
        return response["joints"]

    # -- Useful functions
    def __args_pose_to_list(self, *args):
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, PoseObject):
                return arg.to_list()
            else:
                pose_list = arg
        else:
            pose_list = args

        pose_list_float = self._map_list(pose_list, float)
        if len(pose_list_float) != 6:
            self._raise_exception("A pose should contain 6 elements (x, y, z, roll, pitch, yaw)")
        return pose_list_float

    def __args_joints_to_list(self, *args):
        """
        Convert args into a list
        Either if args = (1.1,5.6,-6.7) or args = ([1.1,5.6,-6.7],) , the
        function will return (1.1,5.6,-6.7)

        :param args: Union[list, tuple]
        :return: list of float
        """
        if len(args) == 1:
            args = args[0]

        joints = self._map_list(args, float)
        if len(joints) != 6:
            self._raise_exception("The robot has 6 joints")

        return joints



import roslibpy
from .enums import ConveyorStatus, ConveyorID, ConveyorDirection


class ConveyorServices(object):

    def __init__(self, client):
        self.__client = client

        self.ping_and_set_conveyor_service = roslibpy.Service(self.__client,
                                                              '/niryo_robot/conveyor/ping_and_set_conveyor',
                                                              'conveyor_interface/SetConveyor')

        self.control_conveyor_service = roslibpy.Service(self.__client,
                                                         '/niryo_robot/conveyor/control_conveyor',
                                                         'conveyor_interface/ControlConveyor')

    @staticmethod
    def get_ping_and_set_conveyor_request():
        return roslibpy.ServiceRequest({"cmd": ConveyorStatus.ADD.value})

    @staticmethod
    def unset_conveyor_request(conveyor_id):
        """

        :param conveyor_id:
        :type conveyor_id: ConveyorID
        :return:
        :rtype: ServiceRequest
        """
        return roslibpy.ServiceRequest({"cmd": ConveyorStatus.REMOVE.value, "id": conveyor_id.value})

    @staticmethod
    def control_conveyor_request(conveyor_id, control_on, speed, direction):
        """

        :param conveyor_id:
        :type conveyor_id: ConveyorID
        :param control_on:
        :type control_on: bool
        :param speed:
        :type speed: int
        :param direction:
        :type direction: ConveyorDirection
        :return:
        :rtype: ServiceRequest
        """
        return roslibpy.ServiceRequest(
            {"id": conveyor_id.value, "control_on": control_on, "speed": speed, "direction": direction.value})



from .niryo_ros import NiryoRos
from .niryo_robot import NiryoRobot
from .niryo_topic import NiryoTopic

from .arm.arm import Arm
from .conveyor.conveyor import Conveyor
from .io.io import IO
from .pick_place.pick_place import PickPlace
from .saved_poses.saved_poses import SavedPoses
from .tool.tool import Tool
from .trajectories.trajectories import Trajectories
from .vision.vision import Vision
from .led_ring.led_ring import LedRing
from .sound.sound import Sound
from .frames.frames import Frames

from .enums import *
from .arm.enums import *
from .conveyor.enums import *
from .io.enums import *
from .saved_poses.enums import *
from .tool.enums import *
from .trajectories.enums import *
from .vision.enums import *
from .led_ring.enums import *
from .sound.enums import *
from .frames.enums import *

from .objects import PoseObject
from .arm.objects import HardwareStatusObject, JointStateObject
from .conveyor.objects import ConveyorInfo
from .io.objects import DigitalPinObject
from .vision.objects import CameraInfo, ImageParameters

import warnings

warnings.warn(
    ('This project is officially marked as deprecated and will no longer be maintained or receive updates starting '
     '1st June 2025. We recommend users migrate to PyNiryo: https://pypi.org/project/pyniryo/'),
    DeprecationWarning,
    stacklevel=2
)



from enum import Enum, unique


@unique
class ToolID(Enum):
    """
    Enumeration of Tools IDs
    """
    NONE = 0
    GRIPPER_1 = 11
    GRIPPER_2 = 12
    GRIPPER_3 = 13
    GRIPPER_4 = 14
    ELECTROMAGNET_1 = 30
    VACUUM_PUMP_1 = 31
    VACUUM_PUMP_2 = 32


@unique
class ToolCommand(Enum):
    # Gripper
    OPEN_GRIPPER = 1
    CLOSE_GRIPPER = 2

    # Vacuump pump
    PULL_AIR_VACUUM_PUMP = 10
    PUSH_AIR_VACUUM_PUMP = 11

    # Tools controlled by digital I/Os
    SETUP_DIGITAL_IO = 20
    ACTIVATE_DIGITAL_IO = 21
    DEACTIVATE_DIGITAL_IO = 22



from pyniryo2.niryo_topic import NiryoTopic
from pyniryo2.arm.objects import HardwareStatusObject, JointStateObject
from pyniryo2.objects import PoseObject


class ArmTopics(object):

    def __init__(self, client):
        self.__client = client

        self.joint_states_topic = NiryoTopic(self.__client,
                                             '/joint_states',
                                             'sensor_msgs/JointState',
                                             joint_states_topic_conversion)

        self.robot_state_topic = NiryoTopic(self.__client,
                                            '/niryo_robot/robot_state',
                                            'niryo_robot_msgs/RobotState',
                                            robot_state_topic_conversion)

        self.hardware_status_topic = NiryoTopic(self.__client,
                                                '/niryo_robot_hardware_interface/hardware_status',
                                                'niryo_robot_msgs/HardwareStatus',
                                                hardware_status_topic_conversion)

        self.learning_mode_state_topic = NiryoTopic(self.__client,
                                                    '/niryo_robot/learning_mode/state',
                                                    'std_msgs/Bool',
                                                    learning_mode_state_topic_conversion)

        self.max_velocity_scaling_factor_topic = NiryoTopic(self.__client,
                                                            '/niryo_robot/max_velocity_scaling_factor',
                                                            'std_msgs/Int32',
                                                            max_velocity_scaling_factor_topic_conversion)


def joint_states_topic_conversion(msg):
    joint_state = JointStateObject()
    joint_state.init_from_message(msg)
    return joint_state


def robot_state_topic_conversion(msg):
    pose = PoseObject(x=msg["position"]["x"],
                      y=msg["position"]["y"],
                      z=msg["position"]["z"],
                      roll=msg["rpy"]["roll"],
                      pitch=msg["rpy"]["pitch"],
                      yaw=msg["rpy"]["yaw"])
    return pose


def hardware_status_topic_conversion(msg):
    hardware_status = HardwareStatusObject()
    hardware_status.init_from_message(msg)
    return hardware_status


def learning_mode_state_topic_conversion(msg):
    return msg["data"]


def max_velocity_scaling_factor_topic_conversion(msg):
    return msg["data"]



import roslibpy

from .enums import ManageSound


class SoundServices(object):

    def __init__(self, client):
        self.__client = client

        self.play_sound_service = roslibpy.Service(self.__client,
                                                   '/niryo_robot_sound/play',
                                                   'niryo_robot_sound/PlaySound')

        self.stop_sound_service = roslibpy.Service(self.__client,
                                                   '/niryo_robot_sound/stop',
                                                   'niryo_robot_msgs/Trigger')

        self.set_sound_volume_service = roslibpy.Service(self.__client,
                                                         '/niryo_robot_sound/set_volume',
                                                         'niryo_robot_msgs/SetInt')

        self.manage_sound_service = roslibpy.Service(self.__client,
                                                     '/niryo_robot_sound/manage',
                                                     'niryo_robot_msgs/ManageSound')

        self.tts_service = roslibpy.Service(self.__client,
                                            '/niryo_robot_sound/text_to_speech',
                                            'niryo_robot_sound/TextToSpeech')

    @staticmethod
    def play_sound_request(sound_name, start_time_sec=0, end_time_sec=0, wait_end=False):
        """

        :param sound_name:
        :type sound_name: str
        param start_time_sec:
        :type start_time_sec: float
        :param end_time_sec:
        :type end_time_sec: float
        :param wait_end:
        :type wait_end: bool
        :return:
        :rtype: ServiceRequest
        """
        return roslibpy.ServiceRequest({"sound_name": sound_name, "start_time_sec": start_time_sec,
                                        "end_time_sec": end_time_sec, "wait_end": wait_end})

    @staticmethod
    def stop_sound_request():
        """

        :return:
        :rtype: ServiceRequest
        """
        return roslibpy.ServiceRequest()

    @staticmethod
    def set_sound_volume_request(value):
        """

        :param value:
        :type value: int8
        :return:
        :rtype: ServiceRequest
        """
        return roslibpy.ServiceRequest({"value": value})

    @staticmethod
    def import_sound_request(sound_name, sound_data):
        """

        :param sound_name:
        :type sound_name: str
        :param sound_data:
        :type sound_data: str
        :return:
        :rtype: ServiceRequest
        """
        return roslibpy.ServiceRequest(
            {"sound_name": sound_name, "sound_data": sound_data, "action": ManageSound.ADD.value})

    @staticmethod
    def delete_sound_request(sound_name):
        """

        :param sound_name:
        :type sound_name: str
        :return:
        :rtype: ServiceRequest
        """
        return roslibpy.ServiceRequest(
            {"sound_name": sound_name, "action": ManageSound.DELETE.value})

    @staticmethod
    def tts_request(text, language):
        """

        :param text:
        :type text: str
        :param language:
        :type language: Language
        :return:
        :rtype: ServiceRequest
        """
        return roslibpy.ServiceRequest({"text": text, "language": language.value})



from enum import Enum, unique


@unique
class ManageTrajectories(Enum):
    """
    Enumeration of actions available for saved trajectories management
    """
    DELETE_ALL = -2
    DELETE = -1
    SAVE = 1
    SAVE_LAST_LEARNED = 2
    UPDATE = 3
    EXECUTE = 4
    EXECUTE_REGISTERED = 5



#!/usr/bin/env python
import unittest
import time
import random

from pyniryo2.exceptions import RobotCommandException
from pyniryo2.niryo_ros import NiryoRos
from pyniryo2.sound.enums import Language
from pyniryo2.sound.sound import Sound
from pyniryo2.niryo_topic import NiryoTopic

robot_ip_address = "127.0.0.1"
port = 9090

test_order = ["test_volume",
              "test_sound",
              "test_say"]


class BaseTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.client = NiryoRos(ip_address=robot_ip_address, port=port)
        cls.sound = Sound(cls.client)

    @classmethod
    def tearDownClass(cls):
        cls.client.close()


# noinspection PyTypeChecker
class TestSound(BaseTest):
    def test_sound(self):
        self.assertIsInstance(self.sound.sounds, list)
        self.assertIsNotNone(self.sound.get_sounds())
        self.assertEqual(self.sound.sounds, self.sound.get_sounds())

        sound_names = self.sound.sounds
        random_sound_name = sound_names[random.randint(0,len(sound_names)-1)]
        self.assertTrue(0 < self.sound.get_sound_duration(random_sound_name))

        start_time = time.time()
        self.assertIsNone(self.sound.play(random_sound_name, True, 0.2, 1.2))
        self.assertTrue(0.5 <= time.time() - start_time <= 3.5)

        start_time = time.time()
        self.assertIsNone(self.sound.play(random_sound_name, False))
        self.assertTrue(time.time() - start_time <= 0.5)

        self.assertEqual(self.sound.state(), random_sound_name)
        self.assertIsNone(self.sound.stop())
        time.sleep(0.1)
        self.assertNotEqual(self.sound.state(), random_sound_name)

        with self.assertRaises(RobotCommandException):
            self.sound.play(450)

        with self.assertRaises(RobotCommandException):
            self.sound.play(True)

    def test_say(self):
        for language in Language:
            start_time = time.time()
            self.assertIsNone(self.sound.say('This is a test', language))
            self.assertTrue(time.time() - start_time > 0.5)

        with self.assertRaises(RobotCommandException):
            self.sound.say('This is a test', 0)

        with self.assertRaises(RobotCommandException):
            self.sound.say(450, Language.ENGLISH)

    def test_volume(self):
        self.assertIsInstance(self.sound.volume, NiryoTopic)
        self.assertIsNotNone(self.sound.volume.value)
        self.assertIsNotNone(self.sound.get_volume())
        self.assertEqual(self.sound.get_volume(), self.sound.volume())

        self.assertIsNone(self.sound.set_volume(50))
        self.assertEqual(self.sound.volume(), 50)

        self.assertIsNone(self.sound.set_volume(10))
        self.assertEqual(self.sound.volume(), 10)

        with self.assertRaises(RobotCommandException):
            self.sound.set_volume(-100)

        with self.assertRaises(RobotCommandException):
            self.sound.set_volume(300)

def suite():
    suite = unittest.TestSuite()
    for function_name in test_order:
        suite.addTest(TestSound(function_name))
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



from pyniryo2.niryo_topic import NiryoTopic

from .objects import ConveyorInfo
from .enums import ConveyorID, ConveyorTTL, ConveyorCan, ConveyorDirection


class ConveyorTopics(object):

    def __init__(self, client):
        self.__client = client

        self.conveyor_feedback_topic = NiryoTopic(self.__client,
                                                  '/niryo_robot/conveyor/feedback',
                                                  'conveyor_interface/ConveyorFeedbackArray',
                                                  conveyor_feedback_topic_conversion)


def conveyor_feedback_topic_conversion(msg):
    return [ConveyorInfo(conveyor_id=conveyor_id_to_conveyor_type(conveyor['conveyor_id']),
                         running=conveyor['running'],
                         speed=conveyor['speed'],
                         direction=ConveyorDirection(conveyor['direction'])) for conveyor in msg['conveyors']]


def conveyor_id_to_conveyor_type(conveyor_id):
    try:
        return ConveyorID(conveyor_id)
    except ValueError:
        try:
            return ConveyorTTL(conveyor_id)
        except ValueError:
            try:
                return ConveyorCan(conveyor_id)
            except ValueError:
                return conveyor_id



"""
Little Script which shows how to control Ned's Led Ring with PyNiryo2
"""
# !/usr/bin/env python

from pyniryo2 import NiryoRobot
import sys, time

if sys.version_info[0] == 2:
    input_func = raw_input
else:
    input_func = input

simulation = "-simu" in sys.argv
robot_ip_address_rpi = "10.10.10.10"
robot_ip_address_simu = "127.0.0.1"
robot_ip_address = robot_ip_address_simu if simulation else robot_ip_address_rpi

# Connecting to robot
niryo_robot = NiryoRobot(ip_address=robot_ip_address)

print('SOLID')
niryo_robot.led_ring.solid([255, 0, 0])

time.sleep(4)

print('FLASH 8 Hz 10 times and wait for the end')
niryo_robot.led_ring.flash([20, 255, 78], iterations=10, wait=True, frequency=8)

print('ALTERNATE 8 times')
niryo_robot.led_ring.alternate([[0, 255, 0], [255, 0, 0], [0, 0, 255]], iterations=8)

time.sleep(4)

print('CHASE with 100ms between each step ')
niryo_robot.led_ring.chase([178, 78, 100], speed=100)

time.sleep(4)

print('WIPE 3 times with 90ms between each step, wait for the end')
niryo_robot.led_ring.wipe([98, 78, 190], speed=90, wait=True)

print('RAINBOW 2 times with 10ms between each step ')
niryo_robot.led_ring.rainbow(speed=10, iterations=2)

time.sleep(4)

print('RAINBOW CYCLE endless')
niryo_robot.led_ring.rainbow_cycle()

time.sleep(4)

print('RAINBOW CHASE endless')
niryo_robot.led_ring.rainbow_chase()

time.sleep(4)

print('GO UP 4 times with 70ms between each step, wait for the end')
niryo_robot.led_ring.go_up([189, 96, 113], iterations=4, speed=70, wait=True)

print('GO UP AND DOWN 5 times, wait for the end')
niryo_robot.led_ring.go_up_down([70, 167, 97], iterations=5, wait=True)

time.sleep(4)

print('Turn off Leds')
niryo_robot.led_ring.turn_off()

# Close
niryo_robot.end()



#!/usr/bin/env python
import unittest
import roslibpy

from pyniryo2.exceptions import RobotCommandException
from pyniryo2.objects import PoseObject
from pyniryo2.niryo_ros import NiryoRos

from pyniryo2.saved_poses.saved_poses import SavedPoses

robot_ip_address = "127.0.0.1"
port = 9090

test_order = ["test_get_saved_poses_list",
              "test_get_saved_pose",
              "test_manage_pose",
              ]


class BaseTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.client = NiryoRos(ip_address=robot_ip_address, port=port)
        cls.saved_poses = SavedPoses(cls.client)

    @classmethod
    def tearDownClass(cls):
        cls.client.close()


# noinspection PyTypeChecker
class TestSavedPoses(BaseTest):
    def test_get_saved_poses_list(self):
        self.assertIsInstance(self.saved_poses.get_saved_pose_list(), list)

    def test_get_saved_pose(self):
        for pose_name in self.saved_poses.get_saved_pose_list():
            self.assertIsInstance(self.saved_poses.get_pose_saved(pose_name), PoseObject)

        if "unittest_pose" in self.saved_poses.get_saved_pose_list():
            self.assertIsNone(self.saved_poses.delete_pose("unittest_pose"))
            self.assertFalse("unittest_pose" in self.saved_poses.get_saved_pose_list())
        with self.assertRaises(RobotCommandException):
            self.saved_poses.get_pose_saved("unittest_pose")

    def test_manage_pose(self):
        if "unittest_pose" in self.saved_poses.get_saved_pose_list():
            self.assertIsNone(self.saved_poses.delete_pose("unittest_pose"))
            self.assertFalse("unittest_pose" in self.saved_poses.get_saved_pose_list())

        self.assertIsNone(self.saved_poses.save_pose("unittest_pose", 0.3, 0.0, 0.3, 0.0, 1.57, 0.0))
        self.assertTrue("unittest_pose" in self.saved_poses.get_saved_pose_list())
        self.assertIsInstance(self.saved_poses.get_pose_saved("unittest_pose"), PoseObject)
        self.assertIsInstance(self.saved_poses.get_saved_pose_list()[0], str)
        self.assertIsNone(self.saved_poses.delete_pose("unittest_pose"))
        self.assertFalse("unittest_pose" in self.saved_poses.get_saved_pose_list())

        self.assertIsNone(self.saved_poses.save_pose("unittest_pose", [0.3, 0.0, 0.3, 0.0, 1.57, 0.0]))
        self.assertTrue("unittest_pose" in self.saved_poses.get_saved_pose_list())
        self.assertIsNone(self.saved_poses.delete_pose("unittest_pose"))
        self.assertFalse("unittest_pose" in self.saved_poses.get_saved_pose_list())

        self.assertIsNone(self.saved_poses.save_pose("unittest_pose", PoseObject(0.3, 0.0, 0.3, 0.0, 1.57, 0.0)))
        self.assertIsNone(self.saved_poses.save_pose("unittest_pose", PoseObject(0.3, 0.0, 0.3, 0.0, 1.57, 0.0)))

        self.assertTrue("unittest_pose" in self.saved_poses.get_saved_pose_list())
        self.assertIsNone(self.saved_poses.delete_pose("unittest_pose"))
        self.assertFalse("unittest_pose" in self.saved_poses.get_saved_pose_list())

        with self.assertRaises(RobotCommandException):
            self.assertIsNone(self.saved_poses.save_pose("unittest_pose", 0.3, 0.0))
        with self.assertRaises(RobotCommandException):
            self.assertIsNone(self.saved_poses.save_pose("unittest_pose",  0.3, 0.0, 0.3, 0.0, 1.57, 0.0, 0.0))
        with self.assertRaises(RobotCommandException):
            self.assertIsNone(self.saved_poses.save_pose("unittest_pose", [0.1]*8))
        with self.assertRaises(RobotCommandException):
            self.assertIsNone(self.saved_poses.save_pose("unittest_pose", [0.1]*4))

        self.assertFalse("unittest_pose" in self.saved_poses.get_saved_pose_list())


def suite():
    suite = unittest.TestSuite()
    for function_name in test_order:
        suite.addTest(TestSavedPoses(function_name))
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



import roslibpy

from pyniryo2.utils import pose_list_to_dict


class ArmServices(object):

    def __init__(self, client):
        self.__client = client

        self.request_calibration_service = roslibpy.Service(self.__client,
                                                            '/niryo_robot/joints_interface/calibrate_motors',
                                                            'niryo_robot_msgs/SetInt')

        self.request_new_calibration_service = roslibpy.Service(self.__client,
                                                                '/niryo_robot/joints_interface/request_new_calibration',
                                                                'niryo_robot_msgs/Trigger')

        self.activate_learning_mode_service = roslibpy.Service(self.__client,
                                                               '/niryo_robot/learning_mode/activate',
                                                               'niryo_robot_msgs/SetBool')

        self.forward_kinematics_service = roslibpy.Service(self.__client,
                                                           '/niryo_robot/kinematics/forward',
                                                           'niryo_robot_arm_commander/GetFK')

        self.inverse_kinematics_service = roslibpy.Service(self.__client,
                                                           '/niryo_robot/kinematics/inverse',
                                                           'niryo_robot_arm_commander/GetIK')

        self.set_max_velocity_scaling_factor_service = roslibpy.Service(
            self.__client,
            '/niryo_robot_arm_commander/set_max_velocity_scaling_factor',
            'niryo_robot_msgs/SetInt'
        )

        self.enable_jog_controller_service = roslibpy.Service(self.__client,
                                                              '/niryo_robot/jog_interface/enable',
                                                              'niryo_robot_msgs/SetBool')

        self.jog_shift_service = roslibpy.Service(self.__client,
                                                  '/niryo_robot/jog_interface/jog_shift_commander',
                                                  'niryo_robot_arm_commander/JogShift')

        self.stop_arm_service = roslibpy.Service(self.__client,
                                                 '/niryo_robot_arm_commander/stop_command',
                                                 'niryo_robot_msgs/Trigger')

        self.get_transform_pose_service = roslibpy.Service(self.__client,
                                                           '/niryo_robot_poses_handlers/get_transform_pose',
                                                           'niryo_robot_msgs/GetTransformPose')

    @staticmethod
    def get_trigger_request():
        return roslibpy.ServiceRequest()

    @staticmethod
    def get_learning_mode_request(enabled):
        return roslibpy.ServiceRequest({"value": enabled})

    @staticmethod
    def get_max_velocity_scaling_factor_request(percentage):
        return roslibpy.ServiceRequest({"value": percentage})

    @staticmethod
    def get_enable_jog_request(enabled):
        return roslibpy.ServiceRequest({"value": enabled})

    @staticmethod
    def get_jog_request(cmd_type, shift_values):
        return roslibpy.ServiceRequest({"cmd": cmd_type, "shift_values": shift_values})

    @staticmethod
    def get_forward_kinematics_request(joint_list):
        return roslibpy.ServiceRequest({"joints": joint_list})

    @staticmethod
    def get_inverse_kinematics_request(pose_list):
        pose_msg = pose_list_to_dict(pose_list)
        return roslibpy.ServiceRequest({"pose": pose_msg})

    @staticmethod
    def get_transform_pose_request(source_frame, local_frame, position, rpy):
        return roslibpy.ServiceRequest({"source_frame": source_frame, "local_frame": local_frame, "position": position, "rpy": rpy})



# - Imports
from __future__ import print_function

# Python libraries
import time

# Communication imports
from .arm.arm import Arm
from .conveyor.conveyor import Conveyor
from .io.io import IO
from .pick_place.pick_place import PickPlace
from .saved_poses.saved_poses import SavedPoses
from .tool.tool import Tool
from .trajectories.trajectories import Trajectories
from .vision.vision import Vision
from .led_ring.led_ring import LedRing
from .sound.sound import Sound
from .frames.frames import Frames
from .niryo_ros import NiryoRos


class NiryoRobot(object):
    def __init__(self, ip_address="127.0.0.1", port=9090):
        """
        Connect your robot to your computer: ::

            robot_simulation = NiryoRobot("127.0.0.1") # Simulation

            robot_hotpot = NiryoRobot("10.10.10.10") # Hotspot

            robot_ethernet = NiryoRobot("169.254.200.201") # Ethernet

        :param ip_address: robot ROS ip
        :type ip_address: str
        :param port: robot ROS port
        :type port: int
        """
        self.__host = ip_address
        self.__port = port
        self.__client = NiryoRos(ip_address, port)

        self.__vision = None
        self.__pick_place = None
        self.__trajectories = None
        self.__tool = None
        self.__sound = None
        self.__saved_poses = None
        self.__io = None
        self.__conveyor = None
        self.__arm = None
        self.__led_ring = None
        self.__frames = None

        self.__arm = Arm(self.__client)
        self.__conveyor = Conveyor(self.__client)
        self.__io = IO(self.__client)
        self.__saved_poses = SavedPoses(self.__client)
        self.__sound = Sound(self.__client)
        self.__tool = Tool(self.__client)
        self.__trajectories = Trajectories(self.__client)
        self.__frames = Frames(self.__client)
        self.__pick_place = PickPlace(self.__client, self.__arm, self.__tool, self.__trajectories)
        self.__vision = Vision(self.__client, self.__arm, self.__tool)
        self.__led_ring = LedRing(self.__client)

    def __del__(self):
        del self.__vision
        del self.__pick_place
        del self.__trajectories
        del self.__tool
        del self.__sound
        del self.__saved_poses
        del self.__io
        del self.__conveyor
        del self.__arm
        del self.__frames

        self.end()

    def __str__(self):
        return "Niryo Robot"

    def __repr__(self):
        return self.__str__()

    @property
    def client(self):
        """
        Get the Niryo Ros client

        :return: NiryoRos client
        :rtype: NiryoRos
        """
        return self.__client

    def end(self):
        """
        Disconnect from your robot and ROS: ::

            # Start
            robot = NiryoRobot("10.10.10.10")

            # End
            robot.end()

        :rtype: None
        """
        if self.__client is not None and self.__client.is_connected:
            self.__client.close()

    @staticmethod
    def wait(duration):
        """
        Wait for a certain time

        :param duration: duration in seconds
        :type duration: float
        :rtype: None
        """
        time.sleep(duration)

    @property
    def arm(self):
        """
        Access to the Arm API

        Example: ::

            robot = NiryoRobot(<robot_ip_address>)
            robot.arm.calibrate_auto()
            robot.arm.move_joints([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])

        :rtype: Arm
        """
        return self.__arm

    @property
    def conveyor(self):
        """
        Access to the Conveyor API

        Example: ::

            robot = NiryoRobot(<robot_ip_address>)
            conveyor_id = robot.conveyor.set_conveyor()
            robot.conveyor.run_conveyor(conveyor_id)

        :rtype: Conveyor
        """
        return self.__conveyor

    @property
    def io(self):
        """
        Access to the I/Os API

        Example: ::

            robot = NiryoRobot(<robot_ip_address>)
            robot.io.set_pin_mode(PinID.GPIO_1A, PinMode.INPUT)
            robot.io.digital_write(PinID.GPIO_1A, PinState.HIGH)

        :rtype: IO
        """
        return self.__io

    @property
    def pick_place(self):
        """
        Access to the PickPlace API

        Example: ::

            robot = NiryoRobot(<robot_ip_address>)
            robot.pick_place.pick_from_pose([0.2, 0.0, 0.1, 0.0, 1.57, 0.0])
            robot.pick_place.place_from_pose([0.0, 0.2, 0.1, 0.0, 1.57, 0.0])

        :rtype: PickPlace
        """
        return self.__pick_place

    @property
    def saved_poses(self):
        """
        Access to the SavedPoses API

        Example: ::

            robot = NiryoRobot(<robot_ip_address>)
            pose_name_list = robot.saved_poses.get_saved_pose_list()
            robot.saved_poses.get_pose_saved(pose_name_list[0])

        :rtype: SavedPoses
        """
        return self.__saved_poses

    @property
    def sound(self):
        """
        Access to the Sound API

        Example: ::

            robot = NiryoRobot(<robot_ip_address>)
            sound.play_sound_user("test_sound.wav")
            sound_name = sound.get_sounds()[0]
            sound_duration = sound.play(sound_name)

        :rtype: Sound
        """
        return self.__sound

    @property
    def tool(self):
        """
        Access to the Tool API

        Example: ::

            robot = NiryoRobot(<robot_ip_address>)
            robot.tool.update_tool()
            robot.tool.grasp_with_tool()
            robot.tool.release_with_tool()

        :rtype: Tool
        """
        return self.__tool

    @property
    def trajectories(self):
        """
        Access to the Trajectories API

        Example: ::

            robot = NiryoRobot(<robot_ip_address>)
            trajectories = robot.trajectories.get_saved_trajectory_list()
            if len(trajectories) > 0:
                robot.trajectories.execute_trajectory_saved(trajectories[0])

        :rtype: Trajectories
        """
        return self.__trajectories

    @property
    def vision(self):
        """
        Access to the Vision API

        Example: ::

            robot = NiryoRobot(<robot_ip_address>)
            robot.vision.vision_pick("workspace_1", 0.0, ObjectShape.ANY, ObjectColor.ANY)
            robot.vision.detect_object("workspace_1", ObjectShape.ANY, ObjectColor.ANY)

        :rtype: Vision
        """
        return self.__vision

    @property
    def led_ring(self):
        """
        Access to the Led Ring API

        Example: ::

            robot = NiryoRobot(<robot_ip_address>)
            niryo_robot.led_ring.led_ring_flash([20,255,78], iterations = 10, wait = True, frequency = 8)
            niryo_robot.led_ring.led_ring_turn_off()

        :rtype: LedRing
        """
        return self.__led_ring

    @property
    def frames(self):
        """
        Access to the frame API

        Example: ::

            robot = NiryoRobot(<robot_ip_address>)

        :rtype: Frames
        """
        return self.__frames



#!/usr/bin/env python
import time
import unittest

from pyniryo2.exceptions import RobotCommandException
from pyniryo2.niryo_ros import NiryoRos

from pyniryo2.conveyor.enums import ConveyorID, ConveyorDirection, ConveyorCan, ConveyorTTL
from pyniryo2.conveyor.conveyor import Conveyor
from pyniryo2.conveyor.topics import ConveyorInfo

robot_ip_address = "127.0.0.1"
port = 9090

test_order = ["test_conveyor_set_run",
              "test_bad_params_errors",
              ]


class BaseTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.client = NiryoRos(ip_address=robot_ip_address, port=port)
        cls.conveyor = Conveyor(cls.client)

    @classmethod
    def tearDownClass(cls):
        cls.client.close()


# noinspection PyTypeChecker
class TestConveyor(BaseTest):
    def test_conveyor_set_run(self):
        conveyor_id = self.conveyor.set_conveyor()
        self.assertIsInstance(conveyor_id, (ConveyorID, ConveyorCan, ConveyorTTL))
        self.assertIsNone(time.sleep(2))

        self.assertIsInstance(self.conveyor.get_conveyors_feedback(), list)
        feedback = self.conveyor.get_conveyors_feedback()
        self.assertIsInstance(feedback[0], ConveyorInfo)
        conveyor_id = feedback[0].conveyor_id
        self.assertTrue(conveyor_id in [ConveyorID.ID_1, ConveyorTTL.ID_1, ConveyorCan.ID_1])

        self.assertIsNone(self.conveyor.control_conveyor(conveyor_id, True, 30, ConveyorDirection.BACKWARD))
        self.assertIsNone(time.sleep(3))
        self.assertEqual(self.conveyor.get_conveyors_feedback()[0], ConveyorInfo(conveyor_id=conveyor_id,
                                                                                 speed=30,
                                                                                 running=True,
                                                                                 direction=ConveyorDirection.BACKWARD))

        self.assertIsNone(self.conveyor.control_conveyor(feedback[0].conveyor_id, True, 50, ConveyorDirection.FORWARD))
        self.assertIsNone(time.sleep(3))
        self.assertEqual(self.conveyor.get_conveyors_feedback()[0], ConveyorInfo(conveyor_id=conveyor_id,
                                                                                 speed=50,
                                                                                 running=True,
                                                                                 direction=ConveyorDirection.FORWARD))

        self.assertIsNone(self.conveyor.control_conveyor(conveyor_id, False, 50, ConveyorDirection.FORWARD))
        self.assertIsNone(time.sleep(3))

        feedback = self.conveyor.get_conveyors_feedback()[0]
        self.assertEqual(feedback.conveyor_id, conveyor_id)
        self.assertEqual(feedback.running, False)

        self.conveyor.run_conveyor(conveyor_id)
        self.assertIsNone(time.sleep(3))
        self.assertEqual(self.conveyor.get_conveyors_feedback()[0], ConveyorInfo(conveyor_id=conveyor_id,
                                                                                 speed=100,
                                                                                 running=True,
                                                                                 direction=ConveyorDirection.FORWARD))

        self.conveyor.stop_conveyor(conveyor_id)
        self.assertIsNone(time.sleep(3))
        self.assertEqual(self.conveyor.get_conveyors_feedback()[0].running, False)
        self.assertEqual(self.conveyor.get_conveyors_feedback()[0].speed, 0)

        self.assertIsNone(self.conveyor.unset_conveyor(conveyor_id))
        self.assertIsNone(time.sleep(1))
        self.assertEqual(self.conveyor.get_conveyors_feedback(), [])
        self.assertFalse(
            conveyor_id in [conveyor.conveyor_id for conveyor in self.conveyor.get_conveyors_feedback()])

        self.assertEqual(self.conveyor.set_conveyor(), conveyor_id)

    def test_bad_params_errors(self):
        self.assertIsNone(time.sleep(1))
        self.assertTrue(self.conveyor.set_conveyor() in [ConveyorID.ID_1, ConveyorTTL.ID_1, ConveyorCan.ID_1])

        with self.assertRaises(RobotCommandException):
            self.conveyor.unset_conveyor(1)

        with self.assertRaises(RobotCommandException):
            self.conveyor.control_conveyor(1, False, 50, ConveyorDirection.FORWARD)

        with self.assertRaises(RobotCommandException):
            self.conveyor.control_conveyor(ConveyorID, False, 50, ConveyorDirection.FORWARD)

        with self.assertRaises(RobotCommandException):
            self.conveyor.control_conveyor(ConveyorID.ID_1, 1, 50, ConveyorDirection.FORWARD)

        with self.assertRaises(RobotCommandException):
            self.conveyor.control_conveyor(ConveyorID.ID_1, True, -100, ConveyorDirection.FORWARD)

        with self.assertRaises(RobotCommandException):
            self.conveyor.control_conveyor(ConveyorID.ID_1, True, 200, ConveyorDirection.FORWARD)

        with self.assertRaises(RobotCommandException):
            self.conveyor.control_conveyor(ConveyorID.ID_1, True, 100, ConveyorDirection)

        with self.assertRaises(RobotCommandException):
            self.conveyor.control_conveyor(ConveyorID.ID_1, True, 100, 1)


def suite():
    suite = unittest.TestSuite()
    for function_name in test_order:
        suite.addTest(TestConveyor(function_name))
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



from enum import Enum, unique


@unique
class ManageFrames(Enum):
    """
    Enumeration of actions available for dynamic frames management
    """
    DELETE = -1
    SAVE = 1
    SAVE_WITH_POINTS = 2
    EDIT = 3



# - Imports
import numpy as np
import functools

from pyniryo2.robot_commander import RobotCommander

from .services import LedRingServices
from .topics import LedRingTopics
from .enums import AnimationMode


def check_ned2_version(func):
    """
    Decorator that check the robot version
    """
    @functools.wraps(func)
    def wrap(*args, **kwargs):
        robot_instance = args[0]
        if robot_instance.client.hardware_version != 'ned2':
            raise Exception("Error Code : BAD_HARDWARE_VERSION\n"
                            "Message : Wrong robot hardware version, feature only available on Ned2")

        return func(*args, **kwargs)

    return wrap


class LedRing(RobotCommander):
    # --- Public functions --- #
    def __init__(self, client):
        """
        LedRing robot functions

        Example: ::

            ros_instance = NiryoRos("10.10.10.10") # Hotspot
            led_ring_interface = LedRing(ros_instance)

        :param client: Niryo ROS client
        :type client: NiryoRos
        """
        super(LedRing, self).__init__(client)

        self._services = LedRingServices(self._client)
        self._topics = LedRingTopics(self._client)
        # self.__action_timeout = 10

    # - Get current status and state of led ring

    @check_ned2_version
    @property
    def status(self):
        """
        Returns the Led Ring status client which can be used synchronously or asynchronously
        to obtain the current Led Ring status (cf LedRingStatusObject).

        Examples: ::

            # Get last value
            led_ring.status()
            led_ring.status.value

            # Subscribe a callback
            def status_callback(msg):
                print([msg.r, msg.g, msg.b])

            led_ring.status.subscribe(status_callback)
            led_ring.status.unsubscribe()

        :return: Led Ring status topic.
        :rtype: NiryoTopic
        """
        return self._topics.led_ring_status_topic

    @check_ned2_version
    def get_status(self):
        """
        Get Led Ring status.

        Example: ::

            status = led_ring.get_status()
            print(status.animation)

        :return: Object with the current led ring mode, the animation played and the color used
        :rtype: LedRingStatusObject
        """
        return self._topics.led_ring_status_topic()

    # - Control Led Ring with available animations
    @check_ned2_version
    def solid(self, color):
        """
        Set the whole Led Ring to a fixed color.

        Example: ::

            led_ring.solid([15, 50, 255])

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :rtype: None
        """
        self.__classic_check_and_execute_w_color(AnimationMode.SOLID, color, 0, 0, True)

    @check_ned2_version
    def turn_off(self):
        """
        Turn off all LEDs

        Example: ::

            led_ring.turn_off()

        :rtype: None
        """
        self.__classic_check_and_execute_without_color(AnimationMode.NONE, 0, 0, True)

    @check_ned2_version
    def flash(self, color, period=0, iterations=0, wait=False, callback=None, timeout=None):
        """
        Flashes a color according to a frequency. The frequency is equal to 1 / period.

        Examples: ::

            # Synchronous use
            led_ring.flash([15, 50, 255])  # Non-blocking
            led_ring.flash([15, 50, 255], 1, 100, False)  # Non-blocking
            led_ring.flash([15, 50, 255], iterations=20, wait=True)  # Wait the end

            frequency = 20  # Hz
            total_duration = 10 # seconds
            led_ring.flash([15, 50, 255], 1./frequency, total_duration * frequency , True)

            # Asynchronous use
            def led_ring_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Failed")
                else:
                    print("Completed with success")

            led_ring.flash([15, 50, 255], iterations=20, wait=True, callback=calibration_callback)


        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive flashes. If 0, the Led Ring flashes endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish all iterations or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """

        self.__classic_check_and_execute_w_color(AnimationMode.FLASHING, color, period, iterations, wait, callback,
                                                 timeout)

    @check_ned2_version
    def alternate(self, color_list, period=0, iterations=0, wait=False, callback=None, timeout=None):
        """
        Several colors are alternated one after the other.

        Examples: ::

            # Synchronous use
            color_list = [
                [15, 50, 255],
                [255, 0, 0],
                [0, 255, 0],
            ]

            led_ring.alternate(color_list) # Non-blocking
            led_ring.alternate(color_list, 1, 100, False) # Non-blocking
            led_ring.alternate(color_list, iterations=20, wait=True) # Blocking

            # Asynchronous use
            def led_ring_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Failed")
                else:
                    print("Completed with success")

            led_ring.alternate(color_list, iterations=20, wait=True, callback=calibration_callback)

        :param color_list: Led color list of lists of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color_list: list[list[float]]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive alternations. If 0, the Led Ring alternates endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish all iterations or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        self.__classic_check_and_execute_w_color_list(AnimationMode.ALTERNATE, color_list, period, iterations, wait,
                                                      callback,
                                                      timeout)

    @check_ned2_version
    def chase(self, color, period=00, iterations=0, wait=False, callback=None, timeout=None):
        """
        Movie theater light style chaser animation.

        Examples: ::

            # Synchronous use
            led_ring.chase([15, 50, 255]) # Non-blocking
            led_ring.chase([15, 50, 255], 1, 100, False) # Non-blocking
            led_ring.chase([15, 50, 255], iterations=20, wait=True) # Blocking

            # Asynchronous use
            def led_ring_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Failed")
                else:
                    print("Completed with success")

            led_ring.chase([15, 50, 255], iterations=20, wait=True, callback=calibration_callback)

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive chase. If 0, the animation continues endlessly.
            One chase just lights one Led every 3 LEDs.
        :type iterations: int
        :param wait: The service wait for the animation to finish all iterations or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        self.__classic_check_and_execute_w_color(AnimationMode.CHASE, color, period, iterations, wait, callback,
                                                 timeout)

    @check_ned2_version
    def wipe(self, color, period=0, wait=False, callback=None, timeout=None):
        """
        Wipe a color across the Led Ring, light a Led at a time.

        Examples: ::

            # Synchronous use
            robot.wipe([15, 50, 255])  # Non-blocking
            led_ring.wipe([15, 50, 255], 1, False)  # Non-blocking
            led_ring.wipe([15, 50, 255], wait=True)  # Blocking

            # Asynchronous use
            def led_ring_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Failed")
                else:
                    print("Completed with success")

            led_ring.wipe([15, 50, 255], wait=True, callback=calibration_callback)

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param wait: The service wait for the animation to finish or not to answer.
        :type wait: bool
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        self.__classic_check_and_execute_w_color(AnimationMode.COLOR_WIPE, color, period, 0, wait, callback,
                                                 timeout)

    @check_ned2_version
    def go_up(self, color, period=0, iterations=0, wait=False, callback=None, timeout=None):
        """
        LEDs turn on like a loading circle, and are then all turned off at once.

        Examples: ::

            # Synchronous use
            led_ring.go_up([15, 50, 255])  # Non-blocking
            led_ring.go_up([15, 50, 255], 1, 100, False)  # Non-blocking
            led_ring.go_up([15, 50, 255], iterations=20, wait=True)  # Blocking

            # Asynchronous use
            def led_ring_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Failed")
                else:
                    print("Completed with success")

            led_ring.go_up([15, 50, 255], period=2, iterations=20, wait=True, callback=calibration_callback)

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive turns around the Led Ring. If 0, the animation
            continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        self.__classic_check_and_execute_w_color(AnimationMode.GO_UP, color, period, iterations, wait, callback,
                                                 timeout)

    @check_ned2_version
    def go_up_down(self, color, period=0, iterations=0, wait=False, callback=None, timeout=None):
        """
        LEDs turn on like a loading circle, and are turned off the same way.

        Examples: ::

            # Synchronous use
            led_ring.go_up_down([15, 50, 255])  # Non-blocking
            led_ring.go_up_down([15, 50, 255], 1, 100, False)  # Non-blocking
            led_ring.go_up_down([15, 50, 255], iterations=20, wait=True)  # Blocking

            # Asynchronous use
            def led_ring_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Failed")
                else:
                    print("Completed with success")

            led_ring.go_up_down([15, 50, 255], period=2, iterations=20, wait=True, callback=calibration_callback)

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive turns around the Led Ring. If 0, the animation
            continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        self.__classic_check_and_execute_w_color(AnimationMode.GO_UP_AND_DOWN, color, period, iterations, wait,
                                                 callback,
                                                 timeout)

    @check_ned2_version
    def breath(self, color, period=0, iterations=0, wait=False, callback=None, timeout=None):
        """
        Variation of the light intensity of the LED ring, similar to human breathing.

        Examples: ::

            # Synchronous use
            led_ring.breath([15, 50, 255])  # Non-blocking
            led_ring.breath([15, 50, 255], 1, 100, False)  # Non-blocking
            led_ring.breath([15, 50, 255], iterations=20, wait=True)  # Blocking

            # Asynchronous use
            def led_ring_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Failed")
                else:
                    print("Completed with success")

            led_ring.breath([15, 50, 255], period=2, iterations=20, wait=True,
                callback=calibration_callback)

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive turns around the Led Ring. If 0, the animation
            continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        self.__classic_check_and_execute_w_color(AnimationMode.BREATH, color, period, iterations, wait, callback,
                                                 timeout)

    @check_ned2_version
    def snake(self, color, period=0, iterations=0, wait=False, callback=None, timeout=None):
        """
        A small coloured snake (certainly a python :D ) runs around the LED ring.

        Examples: ::

            # Synchronous use
            led_ring.snake([15, 50, 255]) # Non-blocking
            led_ring.snake([15, 50, 255], 1, 100, True) # Blocking

            # Asynchronous use
            def led_ring_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Failed")
                else:
                    print("Completed with success")

            led_ring.snake([15, 50, 255], period=2, iterations=20, wait=True, callback=calibration_callback)

        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :param period: Execution time for a pattern in seconds. If 0, the default duration will be used.
        :type period: float
        :param iterations: Number of consecutive turns around the Led Ring. If 0, the animation
            continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        self.__classic_check_and_execute_w_color(AnimationMode.SNAKE, color, period, iterations, wait, callback,
                                                 timeout)

    @check_ned2_version
    def rainbow(self, period=0, iterations=0, wait=False, callback=None, timeout=None):
        """
        Draw rainbow that fades across all LEDs at once.

        Examples: ::

            # Synchronous use
            led_ring.rainbow()  # Non-blocking
            led_ring.rainbow(5, 2, True)  # Blocking
            led_ring.rainbow(wait=True)  # Blocking

            # Asynchronous use
            def led_ring_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Failed")
                else:
                    print("Completed with success")

            led_ring.rainbow(period=2, iterations=20, wait=True, callback=calibration_callback)

        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive rainbows. If 0, the animation continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        self.__classic_check_and_execute_without_color(AnimationMode.RAINBOW, period, iterations, wait, callback,
                                                       timeout)

    @check_ned2_version
    def rainbow_cycle(self, period=0, iterations=0, wait=False, callback=None, timeout=None):
        """
        Draw rainbow that uniformly distributes itself across all LEDs.

        Examples: ::

            # Synchronous use
            led_ring.rainbow_cycle()
            led_ring.rainbow_cycle(5, 2, True)
            led_ring.rainbow_cycle(wait=True)

            # Asynchronous use
            def led_ring_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Failed")
                else:
                    print("Completed with success")

            led_ring.rainbow_cycle(period=2, iterations=20, wait=True, callback=calibration_callback)

        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive rainbow cycles. If 0, the animation continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        self.__classic_check_and_execute_without_color(AnimationMode.RAINBOW_CYLE, period, iterations, wait, callback,
                                                       timeout)

    @check_ned2_version
    def rainbow_chase(self, period=0, iterations=0, wait=False, callback=None, timeout=None):
        """
        Rainbow chase animation, like the led_ring_chase method.

        Examples: ::

            # Synchronous use
            led_ring.rainbow_chase()
            led_ring.rainbow_chase(5, 2, True)
            led_ring.rainbow_chase(wait=True)

            # Asynchronous use
            def led_ring_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Failed")
                else:
                    print("Completed with success")

            led_ring.rainbow_chase(period=2, iterations=20, wait=True, callback=calibration_callback)

        :param period: Execution time for a pattern in seconds. If 0, the default time will be used.
        :type period: float
        :param iterations: Number of consecutive rainbow cycles. If 0, the animation continues endlessly.
        :type iterations: int
        :param wait: The service wait for the animation to finish or not to answer. If iterations
                is 0, the service answers immediately.
        :type wait: bool
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :type timeout: float
        :rtype: None
        """
        self.__classic_check_and_execute_without_color(AnimationMode.RAINBOW_CHASE, period, iterations, wait, callback,
                                                       timeout)

    @check_ned2_version
    def custom(self, led_colors):
        """
        Sends a colour command to all LEDs of the LED ring.
        The function expects a list of colours for the 30 LEDs  of the robot.

        Example: ::

            led_list = [[i / 30. * 255 , 0, 255 - i / 30.] for i in range(30)]
            led_ring.custom(led_list)

            run_flag = True

            def french_flag_moving():
                colors = []
                colors += [[255, 255, 255] for _ in range(2)]
                colors += [[0, 0, 255] for _ in range(11)]
                colors += [[255, 255, 255] for _ in range(4)]
                colors += [[255, 0, 0] for _ in range(11)]
                colors += [[255, 255, 255] for _ in range(2)]

                rate = 10
                while run_flag:
                    for i in range(len(colors)):
                        led_ring.custom(colors[i:] + colors[:i])
                        time.sleep(1/rate)
                        if not run_flag:
                            return

            french_flag_moving()

        :param led_colors: List of size 30 of led color in a list of size 3[R, G, B].
                RGB channels from 0 to 255.
        :type led_colors: list[list[float]]
        :rtype: None
        """
        self._check_length(led_colors, 30)
        self.__classic_check_and_execute_w_color_list(AnimationMode.CUSTOM, led_colors, 0, 0, True)

    def set_led_color(self, led_id, color):
        """
        Lights up an LED in one colour. RGB colour between 0 and 255.

        Example: ::

            robot.set_led_color(5, [15, 50, 255])

        :param led_id: Id of the led: between 0 and 29
        :type led_id: int
        :param color: Led color in a list of size 3[R, G, B]. RGB channels from 0 to 255.
        :type color: list[float]
        :rtype: None
        """
        color = self._check_color(color)
        self._check_type(led_id, int)

        req = self._services.set_led_ring_color_request(led_id, color)
        resp = self._services.set_led_ring_led_color_service.call(req)
        self._check_result_status(resp)

    # Usefull method Led Ring
    def __classic_check_and_execute_w_color(self, animation, color, period, iterations, wait, callback=None,
                                            timeout=None):
        checked_color = self._check_color(color)
        self._check_instance(period, (float, int))
        self._check_type(iterations, int)
        self._check_type(wait, bool)
        self._check_enum_belonging(animation, AnimationMode)

        req = self._services.set_led_ring_request(animation, color_list=checked_color, period=period,
                                                  iterations=iterations, wait=wait)
        resp = self._services.set_led_ring_animation_service.call(req, callback=callback, timeout=timeout)

        if not callback:
            self._check_result_status(resp)

    def __classic_check_and_execute_w_color_list(self, animation, color_list, period, iterations, wait, callback=None,
                                                 timeout=None):
        self._check_instance(period, (float, int))
        self._check_type(iterations, int)
        self._check_type(wait, bool)
        self._check_enum_belonging(animation, AnimationMode)

        checked_color_list = [self._check_color(color) for index, color in enumerate(color_list)]

        req = self._services.set_led_ring_request(animation, color_list=checked_color_list, period=period,
                                                  iterations=iterations, wait=wait)
        resp = self._services.set_led_ring_animation_service.call(req, callback=callback, timeout=timeout)

        if not callback:
            self._check_result_status(resp)

    def __classic_check_and_execute_without_color(self, animation, period, iterations, wait, callback=None,
                                                  timeout=None):
        self._check_instance(period, (float, int))
        self._check_type(iterations, int)
        self._check_type(wait, bool)
        self._check_enum_belonging(animation, AnimationMode)

        req = self._services.set_led_ring_request(animation, period=period, iterations=iterations,
                                                  wait=wait)
        resp = self._services.set_led_ring_animation_service.call(req, callback=callback, timeout=timeout)

        if not callback:
            self._check_result_status(resp)
            
    def _check_color(self, color):
        checked_color = []
        self._check_type(color, list)
        if len(color) != 3:
            self._raise_exception("Color must be a list of size 3: [r, g, b]")
        for color_elem in color:
            if not 0 <= color_elem <= 255:
                self._raise_exception_expected_range(0, 255, color_elem)
            checked_color.append(self._transform_to_type(color_elem, float))
        return checked_color
